###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurOrbControlWorkChain' for finding the groundstate
    in a DFT+U calculation.
"""
from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction as cf
from aiida.orm import Dict, Code, StructureData, RemoteData
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur, get_inputs_inpgen

from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.data.fleurinp import FleurinpData, get_fleurinp_from_remote_data

import numpy as np
from lxml import etree


def generate_density_matrix_configurations(occupations=None, configurations=None):
    """
    Generate all the necessary density matrix configurations from either the occupations or
    the explicitly given configurations for each species/orbital

    Both arguments are expected as dictionaries in the form ``d[species][orbital]``, with the
    orbital key holding the specification for the current LDA+U procedure

    :param occupations: specifying the occupations for each procedure
    :param configurations: specifying a explicit list of configurations that should be calculated


    :returns: list of dictionaries with all the possible starting configurations for the whole system
    """
    from more_itertools import distinct_permutations
    from itertools import product

    if occupations is not None and configurations is not None:
        raise ValueError('Please provide either occupations or configurations not both')

    config_dict = {}  #This dictionary will contain all distinct permutations for each
    #species/orbital. They are recombined after
    if occupations is not None:
        #The initial occupations were given

        for species, occ_species in occupations.items():
            for orbital, fixed_occ in occ_species.items():
                ind = f'{species}-{orbital}'
                l = int(orbital)
                config_dict[ind] = []

                if not isinstance(fixed_occ, list):
                    spin_occupation = fixed_occ // 2
                    if fixed_occ % 2 == 0:
                        fixed_occ = [spin_occupation, spin_occupation]
                    else:
                        fixed_occ = [spin_occupation + 1, spin_occupation]  #not ideal but for fine for now

                if any(x > 2 * l + 1 for x in fixed_occ):
                    raise ValueError(f'Invalid occupation {species} {orbital}: {fixed_occ}')

                for occ in fixed_occ:
                    spin_configs = []
                    start = [0 for _ in range(2 * l + 1)]
                    #Fill up the occupations until it matches the wanted one
                    i = 0
                    while sum(start) < occ:
                        start[i] = 1
                        i += 1
                    for config in distinct_permutations(start):
                        spin_configs.append(config)

                    if len(config_dict[ind]) != 0:
                        spin_configs = product(config_dict[ind], spin_configs)

                    combined_config = []
                    for config in spin_configs:
                        combined_config.append(config)
                    config_dict[ind] = combined_config

    elif configurations is not None:
        #The configurations were given explicitely

        for species, configs_species in configurations.items():
            for orbital, configs in configs_species.items():
                ind = f'{species}-{orbital}'
                if not isinstance(configs[0], list):
                    configs = [configs]
                config_dict[ind] = configs

    #Now combine them
    all_atom_configs = []
    for configs in product(*config_dict.values()):

        current_config = {}
        for index, key in enumerate(config_dict.keys()):
            current_config[key] = configs[index]
        all_atom_configs.append(current_config)

    return all_atom_configs


class FleurOrbControlWorkChain(WorkChain):
    """
    Workchain for determining the groundstate density matrix in an DFT+U
    calculation. This is done in 2 or 3 steps:

        1. Converge the system without DFT+U (a converged calculation can be
           provided to skip this step)
        2. A fixed number of iterations is run with fixed density matrices
           either generated as all distinct permutations for the given occupations
           or the explicitly given configurations
        3. The system and density matrix is relaxed

    :param wf_parameters: (Dict), Workchain Specifications
    :param scf_no_ldau: (Dict), Inputs to a FleurScfWorkChain providing the initial system
                                either converged or staring from a structure
    :param scf_with_ldau: (Dict), Inputs to a FleurScfWorkChain. Only the wf_parameters are valid
    :param fleurinp: (FleurinpData) FleurinpData to start from if no SCF should be done
    :param remote: (RemoteData) RemoteData to start from if no SCF should be done
    :param structure: (StructureData) Structure to start from if no SCF should be done
    :param calc_parameters: (Dict), Inpgen Parameters
    :param settings: (Dict), additional settings for e.g retrieving files
    :param options: (Dict), Options for the submission of the jobs
    :param inpgen: (Code)
    :param fleur: (Code)
    """
    _workflowversion = '0.3.3'

    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 2 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    _wf_default = {
        'iterations_fixed': 30,
        'ldau_dict': None,
        'use_orbital_occupation': False,
        'fixed_occupations': None,
        'fixed_configurations': None,
        'inpxml_changes': []
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False,
                               'help': 'Inputs for SCF Workchain before adding LDA+U'
                           },
                           namespace='scf_no_ldau')
        spec.input('remote', valid_type=RemoteData, required=False)
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('structure', valid_type=StructureData, required=False)
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False,
                               'help': 'Inputs for SCF Workchain after the LDA+U matrix was fixed'
                           },
                           exclude=('structure', 'fleurinp', 'remote_data'),
                           namespace='scf_with_ldau')
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('inpgen', valid_type=Code, required=False)
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('options', valid_type=Dict, required=False)
        spec.input('options_inpgen', valid_type=Dict, required=False)
        spec.input('settings', valid_type=Dict, required=False)
        spec.input('settings_inpgen', valid_type=Dict, required=False)

        spec.outline(cls.start, cls.validate_input,
                     if_(cls.scf_no_ldau_needed)(cls.converge_scf_no_ldau).elif_(cls.inpgen_needed)(cls.run_inpgen),
                     cls.create_configurations, cls.run_fleur_fixed, cls.converge_scf, cls.return_results)

        spec.output('output_orbcontrol_wc_para', valid_type=Dict)
        spec.expose_outputs(FleurScfWorkChain, namespace='groundstate_scf')

        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(233,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Input codes do not correspond to fleur or inpgen respectively.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(342,
                       'ERROR_SOME_CONFIGS_FAILED',
                       message='Convergence LDA+U calculation failed for some Initial configurations.')
        spec.exit_code(343,
                       'ERROR_ALL_CONFIGS_FAILED',
                       message='Convergence LDA+U calculation failed for all Initial configurations.')
        spec.exit_code(360, 'ERROR_INPGEN_CALCULATION_FAILED', message='Inpgen calculation failed.')
        spec.exit_code(450, 'ERROR_SCF_NOLDAU_FAILED', message='Convergence workflow without LDA+U failed.')

    def start(self):
        """
        init context and some parameters
        """
        self.report(f'INFO: started orbital occupation control workflow version {self._workflowversion}')

        ####### init    #######

        # internal para /control para
        self.ctx.scf_no_ldau = None
        self.ctx.scf_no_ldau_needed = False
        self.ctx.inpgen_needed = False
        self.ctx.fixed_configurations = []
        self.ctx.successful = True
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.description_wf = self.inputs.get('description', '') + '|fleur_orbcontrol_wc|'
        self.ctx.label_wf = self.inputs.get('label', 'fleur_orbcontrol_wc')

        wf_default = self._wf_default
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        # initialize the dictionary using defaults if no options are given
        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

    def validate_input(self):
        """
        validate input
        """
        extra_keys = []
        for key in self.ctx.wf_dict:
            if key not in self._wf_default:
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for Orbcontrol contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        ldau_dict = self.ctx.wf_dict.get('ldau_dict')
        ldau_keys_required = ['l', 'U', 'J', 'l_amf']
        if ldau_dict is not None:
            missing = []
            for species, current in ldau_dict.items():
                for key in ldau_keys_required:
                    if key not in current:
                        missing.append(key)
                if missing:
                    error = 'ERROR: Missing input: The following required keys are missing from ldau_dict' \
                            f' for species {species}: {missing}'
                    self.report(error)
                    return self.exit_codes.ERROR_INVALID_INPUT_PARAM
        else:
            error = 'ERROR: Missing input: ldau_dict was not speciified'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        max_iters = self.ctx.wf_dict.get('iterations_fixed')
        if max_iters <= 1:
            error = "ERROR: 'iterations_fixed' should be equal at least 2"
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        occupations_dict = self.ctx.wf_dict.get('fixed_occupations')
        configurations_dict = self.ctx.wf_dict.get('fixed_configurations')
        #TODO:check occupations or configurations
        if occupations_dict is not None:
            if configurations_dict is not None:
                error = 'ERROR: Invalid input: Only provide one of fixed_occupations and fixed_configurations'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

            for species, occ_species in occupations_dict.items():
                for orbital, occ in occ_species.items():
                    if species not in ldau_dict:
                        error = f'ERROR: Invalid input: {species} defined in fixed_occupations but not in ldau_dict'
                        self.report(error)
                        return self.exit_codes.ERROR_INVALID_INPUT_PARAM
                    else:
                        missing = False
                        if isinstance(ldau_dict[species], dict):
                            if int(orbital) != ldau_dict[species]['l']:
                                missing = True
                        else:
                            for index, current_dict in enumerate(ldau_dict[species]):
                                if int(orbital) == current_dict['l']:
                                    break
                                if index == len(ldau_dict) - 1:
                                    missing = True
                        if missing:
                            error = f'ERROR: Invalid input: Orbital {orbital} is given in fixed_occupations for {species}, ' \
                                     ' but it is not defined in ldau_dict'
                            self.report(error)
                            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

                    if not isinstance(occ, list):
                        error = f'ERROR: Invalid input: {species} defined in fixed_occupations invalid type'
                        self.report(error)
                        return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        else:
            if configurations_dict is not None:

                for species, occ_species in occupations_dict.items():
                    for orbital, occ in occ_species.items():

                        if species not in ldau_dict:
                            error = f'ERROR: Invalid input: {species} defined in fixed_configurations but not in ldau_dict'
                            self.report(error)
                            return self.exit_codes.ERROR_INVALID_INPUT_PARAM
                        else:
                            missing = False
                            if isinstance(ldau_dict[species], dict):
                                if int(orbital) != ldau_dict[species]['l']:
                                    missing = True
                            else:
                                for index, current_dict in enumerate(ldau_dict[species]):
                                    if int(orbital) == current_dict['l']:
                                        break
                                    if index == len(ldau_dict) - 1:
                                        missing = True
                            if missing:
                                error = f'ERROR: Invalid input: Orbital {orbital} is given in fixed_configurations for {species}, ' \
                                         ' but it is not defined in ldau_dict'
                                self.report(error)
                                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

                        if not isinstance(occ, list):
                            error = f'ERROR: Invalid input: {species} defined in fixed_configurations invalid type'
                            self.report(error)
                            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

            else:
                error = 'ERROR: Missing input: Provide one of fixed_occupations or fixed_configurations'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        inputs = self.inputs
        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for INPGEN does not use the plugin fleur.inpgen')
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        fleurinp = None
        remote = None
        if 'scf_no_ldau' in inputs:
            input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf_no_ldau'))
            self.ctx.scf_no_ldau_needed = True
            if 'fleurinp' in input_scf:
                fleurinp = input_scf.fleurinp
            if 'remote_data' in input_scf:
                remote = input_scf.remote_data
            if 'remote' in inputs:
                error = 'ERROR: you gave SCF input + remote for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'fleurinp' in inputs:
                error = 'ERROR: you gave SCF input + fleurinp for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'structure' in inputs:
                error = 'ERROR: you gave SCF input + structure for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'calc_parameters' in inputs:
                error = 'ERROR: you gave SCF input + calc_parameters for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'inpgen' in inputs:
                error = 'ERROR: you gave SCF input + inpgen for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'structure' in inputs:
            self.ctx.inpgen_needed = True
            if 'inpgen' not in inputs:
                error = 'ERROR: you gave structure input but no inpgen code Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'remote' not in inputs and 'fleurinp' not in inputs:
            error = 'ERROR: you gave neither SCF input nor remote or fleurinp'
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            if 'calc_parameters' in inputs:
                error = 'ERROR: you gave remote/fleurinp input + calc_parameters for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'structure' in inputs:
                error = 'ERROR: you gave remote/fleurinp input + structure for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'inpgen' in inputs:
                error = 'ERROR: you gave remote/fleurinp input + inpgen for the Orbcontrol calculation'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'remote' in inputs:
                remote = inputs.remote
            if 'fleurinp' in inputs:
                fleurinp = inputs.fleurinp

        if fleurinp is not None:
            modes = fleurinp.get_fleur_modes()
            if modes['ldau']:
                error = f"ERROR: Wrong input: fleurinp {'in scf_no_ldau' if 'scf_no_ldau' in inputs else ''} already contains LDA+U"
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        if remote is not None:
            retrieved_filenames = remote.creator.outputs.retrieved.list_object_names()
            if FleurCalculation._NMMPMAT_FILE_NAME in retrieved_filenames or \
               FleurCalculation._NMMPMAT_HDF5_FILE_NAME in retrieved_filenames:
                error = f"ERROR: Wrong input: remote_data {'in scf_no_ldau' if 'scf_no_ldau' in inputs else ''} already contains LDA+U"
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

    def scf_no_ldau_needed(self):
        """
        Returns whether to run an additional scf workchain before adding LDA+U
        """
        return self.ctx.scf_no_ldau_needed

    def converge_scf_no_ldau(self):
        """
        Launch fleur.scf for the system without LDA+U
        """
        inputs = self.get_inputs_scf_no_ldau()

        self.report('Info: Run SCF without LDA+U')
        future = self.submit(FleurScfWorkChain, **inputs)
        future.label = 'scf_no_ldau'
        future.description = 'SCF Calculation for orbital occupation control before adding DFT+U'
        return ToContext(scf_no_ldau=future)

    def get_inputs_scf_no_ldau(self):
        """
        Get the inputs for the scf workchain without LDA+U
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf_no_ldau'))

        if 'fleur' not in input_scf:
            input_scf.fleur = self.inputs.fleur

        if 'options' not in input_scf:
            input_scf.options = self.inputs.options

        input_scf.metadata.call_link_label = 'scf_no_ldau'
        return input_scf

    def inpgen_needed(self):
        """
        Returns whether the inpgen should be run directly by this workchain
        """
        return self.ctx.inpgen_needed

    def run_inpgen(self):
        """
        Run the input generator
        """
        ## prepare inputs for inpgen
        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        label = 'scf: inpgen'
        description = f'{self.ctx.description_wf} inpgen on {self.ctx.formula}'

        inpgencode = self.inputs.inpgen

        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None

        if 'settings_inpgen' in self.inputs:
            settings = self.inputs.settings_inpgen
        else:
            settings = None

        if 'options_inpgen' in self.inputs:
            options = self.inputs.options_inpgen
        else:
            #Only take the parts that could be relevant (resources is overwritten anyway)
            options = {'queue_name': self.inputs.options.get_dict().get('queue_name', '')}
            if 'max_wallclock_seconds' in options:
                options['max_wallclock_seconds'] = int(self.inputs.options['max_wallclock_seconds'])

        inputs_build = get_inputs_inpgen(structure,
                                         inpgencode,
                                         options,
                                         label,
                                         description,
                                         settings=settings,
                                         params=params)

        inputs_build.metadata.call_link_label = 'inpgen'
        # Launch inpgen
        self.report('INFO: run inpgen')
        future = self.submit(inputs_build)
        future.label = 'inpgen'
        future.description = 'Inpgen calculation for Orbital occupation control workflow'

        return ToContext(inpgen=future)

    def create_configurations(self):
        """
        Creates the configurations for the initial density matrices

        If fixed_occupations was provided the density matrices are constructed
        as having the given occupations and constructing all distinct permutations

        If fixed_configurations was provided only the given configurations are taken
        """
        all_atom_configs = generate_density_matrix_configurations(
            occupations=self.ctx.wf_dict['fixed_occupations'], configurations=self.ctx.wf_dict['fixed_configurations'])

        self.report(f'INFO: Generated {len(all_atom_configs)} distinct fixed starting configurations')
        self.ctx.fixed_configurations = all_atom_configs

    def run_fleur_fixed(self):
        """
        Launches fleur.base with l_linMix=T and mixParam=0.0, i.e. with a fixed density matrix
        for all configurations.
        """

        if self.ctx.scf_no_ldau_needed:

            if not self.ctx.scf_no_ldau.is_finished_ok:
                error = ('ERROR: SCF workflow without LDA+U was not successful')
                self.report(error)
                return self.exit_codes.ERROR_SCF_NOLDAU_FAILED

            try:
                self.ctx.scf_no_ldau.outputs.output_scf_wc_para
            except NotExistent:
                message = ('ERROR: SCF workflow without LDA+U failed, no scf output node')
                self.ctx.errors.append(message)
                return self.exit_codes.ERROR_SCF_NOLDAU_FAILED

        self.report('INFO: Run fixed density matrix calculations')

        for index, config in enumerate(self.ctx.fixed_configurations):

            inputs, status = self.get_inputs_fixed_configurations(index, config)
            if status:
                return status
            label = f'Fixed_{index}'

            inputs.setdefault('metadata', {})['call_link_label'] = label
            res = self.submit(FleurBaseWorkChain, **inputs)
            res.label = label
            res.description = f'DFT+U calculation with fixed configuration number {index}'
            self.to_context(**{label: res})

    def get_inputs_fixed_configurations(self, index, config):
        """
        Sets up the input for the fixed density matrix calculation.
        """

        remote_data = None
        if self.ctx.scf_no_ldau_needed:
            try:
                fleurinp = self.ctx.scf_no_ldau.outputs.fleurinp
                remote_data = self.ctx.scf_no_ldau.outputs.last_calc.remote_folder
            except NotExistent:
                error = 'Fleurinp generated in the SCF calculation is not found.'
                self.control_end_wc(error)
                return {}, self.exit_codes.ERROR_SCF_NOLDAU_FAILED
        elif self.ctx.inpgen_needed:
            if not self.ctx.inpgen.is_finished_ok:
                error = 'Inpgen calculation failed'
                self.control_end_wc(error)
                return {}, self.exit_codes.ERROR_INPGEN_CALCULATION_FAILED
            try:
                fleurinp = self.ctx.inpgen.outputs.fleurinpData
            except (AttributeError, NotExistent):
                return {}, self.exit_codes.ERROR_INPGEN_CALCULATION_FAILED
        else:
            if 'remote' in self.inputs:
                remote_data = self.inputs.remote
            if 'fleurinp' not in self.inputs:
                fleurinp = get_fleurinp_from_remote_data(remote_data, store=True)
                self.report(
                    f'INFO: generated FleurinpData from {fleurinp.files} from remote folder pk={remote_data.pk}')
            else:
                fleurinp = self.inputs.fleurinp

        inputs = self.inputs

        label = f'Fixed_{index}'
        description = f'LDA+U with fixed nmmpmat for config {index}'

        if 'settings' not in inputs:
            settings = {}
        else:
            settings = inputs.settings.get_dict()
        settings.setdefault('remove_from_remotecopy_list', []).append('mixing_history*')

        self.report(f'INFO: create fleurinp for config {index}')
        fm = FleurinpModifier(fleurinp)
        modes = fleurinp.get_fleur_modes()

        fm.set_inpchanges({'itmax': self.ctx.wf_dict['iterations_fixed'], 'l_linMix': True, 'mixParam': 0.0})

        for atom_species, ldau_dict in self.ctx.wf_dict['ldau_dict'].items():
            fm.set_species(species_name=atom_species, attributedict={'ldaU': ldau_dict})

        for config_index, config_species in config.items():
            orbital = config_index.split('-')[-1]
            atom_species = '-'.join(config_index.split('-')[:-1])

            if len(config_species) == 2 and modes['jspin'] == 1:
                self.report(f'Configuration for species {atom_species} is given spin-polarized, '
                            'but the calculation is non-spinpolarized. Summing up configurations.')
                config_species = [sum(np.array(config) for config in config_species).tolist()]

            for spin, config_spin in enumerate(config_species):
                if self.ctx.wf_dict['use_orbital_occupation']:
                    fm.set_nmmpmat(species_name=atom_species,
                                   orbital=int(orbital),
                                   spin=spin + 1,
                                   orbital_occupations=config_spin)
                else:
                    fm.set_nmmpmat(species_name=atom_species,
                                   orbital=int(orbital),
                                   spin=spin + 1,
                                   state_occupations=config_spin)

        fchanges = self.ctx.wf_dict['inpxml_changes']
        if fchanges:
            try:
                fm.add_task_list(fchanges)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply inpxml_changes'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return {}, self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        try:
            fm.show(display=False, validate=True)
        except etree.DocumentInvalid:
            self.control_end_wc('ERROR: input, inp.xml changes did not validate')
            return {}, self.exit_codes.ERROR_INVALID_INPUT_FILE
        except ValueError as exc:
            error = ('ERROR: input, inp.xml changes could not be applied.'
                     f'The following error was raised {exc}')
            self.control_end_wc(error)
            return {}, self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        fleurinp_fixed = fm.freeze()

        input_fixed = get_inputs_fleur(inputs.fleur,
                                       remote_data,
                                       fleurinp_fixed,
                                       self.ctx.options,
                                       label,
                                       description,
                                       settings=settings)

        return input_fixed, None

    def converge_scf(self):
        """
        Launch fleur.scf after the fixed density matrix calculations to relax the density matrix
        """
        self.report('INFO: Relax density matrices')
        for index, config in enumerate(self.ctx.fixed_configurations):

            inputs = self.get_inputs_scf()
            fixed_calc = self.ctx[f'Fixed_{index}']

            if not fixed_calc.is_finished_ok:
                message = f'One Base workflow (fixed nmmpmat) failed: {index}'
                self.ctx.warnings.append(message)
                continue

            try:
                fixed_calc.outputs.output_parameters
            except NotExistent:
                message = f'One Base workflow (fixed nmmpmat) failed, no output node: {index}. I skip this one.'
                self.ctx.errors.append(message)
                continue

            inputs.fleurinp = fixed_calc.inputs.fleurinpdata
            inputs.remote_data = fixed_calc.outputs.remote_folder

            label = f'Relaxed_{index}'
            inputs.setdefault('metadata', {})['call_link_label'] = label
            res = self.submit(FleurScfWorkChain, **inputs)
            res.label = label
            res.description = f'DFT+U calculation for configuration number {index} converging the density matrix'
            self.to_context(**{label: res})

    def get_inputs_scf(self):
        """
        Get the input for the scf workchain after the fixed density matrix calculations
        to relax the density matrix
        """
        if 'scf_with_ldau' in self.inputs:
            input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf_with_ldau'))
        else:
            input_scf = AttributeDict({})

        if 'fleur' not in input_scf:
            input_scf.fleur = self.inputs.fleur

        if 'options' not in input_scf:
            input_scf.options = self.inputs.options

        if 'settings' not in input_scf:
            settings = {}
        else:
            settings = input_scf.settings.get_dict()
        settings.setdefault('remove_from_remotecopy_list', []).append('mixing_history*')

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        scf_wf_dict.setdefault('inpxml_changes', []).append(('set_inpchanges', {
            'change_dict': {
                'l_linMix': False,
            }
        }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)
        input_scf.settings = Dict(dict=settings)

        return input_scf

    def return_results(self):
        """
        return the results of the relaxed DFT+U calculations (scf workchains)
        """
        distancelist = []
        t_energylist = []
        failed_configs = []
        non_converged_configs = []
        configs_list = []
        outnodedict = {}
        e_u = 'htr'
        dis_u = 'me/bohr^3'
        for index, config in enumerate(self.ctx.fixed_configurations):
            if f'Relaxed_{index}' in self.ctx:
                calc = self.ctx[f'Relaxed_{index}']
            else:
                message = (f'One SCF workflow was not run because the fixed calculation failed: Relaxed_{index}')
                self.ctx.warnings.append(message)
                self.ctx.successful = False
                failed_configs.append(index)
                t_energylist.append(None)
                distancelist.append(None)
                continue

            if not calc.is_finished_ok:
                message = f'One SCF workflow was not successful: Relaxed_{index}'
                self.ctx.warnings.append(message)
                self.ctx.successful = False
                #We dont skip simply non-converged calculations
                #because we want to try to exctract the total_energy
                if calc.exit_status not in FleurScfWorkChain.get_exit_statuses(['ERROR_DID_NOT_CONVERGE']):
                    failed_configs.append(index)
                    t_energylist.append(None)
                    distancelist.append(None)
                    continue
                non_converged_configs.append(index)

            try:
                outputnode_scf = calc.outputs.output_scf_wc_para
            except NotExistent:
                message = f'One SCF workflow failed, no scf output node: Relaxed_{index}. I skip this one.'
                self.ctx.errors.append(message)
                self.ctx.successful = False
                failed_configs.append(index)
                t_energylist.append(None)
                distancelist.append(None)
                continue

            try:
                fleurinp_scf = calc.outputs.fleurinp
            except NotExistent:
                message = f'One SCF workflow failed, no fleurinp output node: Relaxed_{index}. I skip this one.'
                self.ctx.errors.append(message)
                self.ctx.successful = False
                failed_configs.append(index)
                t_energylist.append(None)
                distancelist.append(None)
                continue

            # we loose the connection of the failed scf here.
            # link labels cannot contain '.'
            link_label = f'configuration_{index}'
            fleurinp_label = f'fleurinp_{index}'
            outnodedict[link_label] = outputnode_scf
            outnodedict[fleurinp_label] = fleurinp_scf

            outpara = outputnode_scf.get_dict()

            t_e = outpara.get('total_energy', None)
            e_u = outpara.get('total_energy_units', 'htr')
            dis = outpara.get('distance_charge', None)
            dis_u = outpara.get('distance_charge_units', 'me/bohr^3')
            t_energylist.append(t_e)
            distancelist.append(dis)
            configs_list.append(index)

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            'configurations': self.ctx.fixed_configurations,
            'total_energy': t_energylist,
            'total_energy_units': e_u,
            'distance_charge': distancelist,
            'distance_charge_units': dis_u,
            'successful_configs': configs_list,
            'non_converged_configs': non_converged_configs,
            'failed_configs': failed_configs,
            'groundstate_configuration': None,
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors
        }

        if self.ctx.successful:
            self.report('Done, Orbital occupation control calculation complete')
        elif len(t_energylist) != 0:
            self.report('Done, but something went wrong.... Probably some individual calculation failed or'
                        ' a scf-cycle did not reach the desired distance.')
        else:
            self.report('Done, but something went wrong.... All Calculations failed. Probably something is'
                        ' wrong in your setup')

        #Find the minimal total energy in the list
        if len(t_energylist) != 0:
            groundstate_index = np.nanargmin(np.array(t_energylist, dtype=np.float))
            out['groundstate_configuration'] = groundstate_index

            if f'Relaxed_{groundstate_index}' in self.ctx:
                groundstate_scf = self.ctx[f'Relaxed_{groundstate_index}']
                self.out_many(self.exposed_outputs(groundstate_scf, FleurScfWorkChain, namespace='groundstate_scf'))

        outnode = Dict(dict=out)
        outnodedict['results_node'] = outnode

        # create links between all these nodes...
        outputnode_dict = create_orbcontrol_result_node(**outnodedict)

        outputnode = outputnode_dict.get('output_orbcontrol_wc_para')
        outputnode.label = 'output_orbcontrol_wc_para'
        outputnode.description = (
            'Contains orbital occupation control results and information of an FleurOrbControlWorkChain run.')

        self.out('output_orbcontrol_wc_para', outputnode)

        outputscf = outputnode_dict.get('output_orbcontrol_wc_gs_scf', None)
        if outputscf:
            outputscf.label = 'output_orbcontrol_wc_gs_scf'
            outputscf.description = ('SCF output from the run with the lowest total '
                                     'energy extracted from FleurOrbControlWorkChain')

        if all(e is None for e in t_energylist):
            return self.exit_codes.ERROR_ALL_CONFIGS_FAILED
        elif not self.ctx.successful:
            return self.exit_codes.ERROR_SOME_CONFIGS_FAILED

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_orbcontrol_result_node(**kwargs):
    """
    This is a pseudo cf, to create the right graph structure of AiiDA.
    This calcfunction will create the output nodes in the database.
    It also connects the output_nodes to all nodes the information comes from.
    This includes the output_parameter node for the orbcontrol, connections to run scfs,
    and returning of the gs_calculation (best initial density matrix)
    So far it is just parsed in as kwargs argument, because we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_orbcontrol_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice
    # then have a circle in the database for now...
    outputdict = outpara.get_dict()
    groundstate_index = outputdict.get('groundstate_configuration')
    if groundstate_index is None:
        return outdict

    if f'configuration_{groundstate_index}' in kwargs:
        outdict['output_orbcontrol_wc_gs_scf'] = kwargs[f'configuration_{groundstate_index}'].clone()
    if f'fleurinp_{groundstate_index}' in kwargs:
        outdict['output_orbcontrol_wc_gs_fleurinp'] = kwargs[f'fleurinp_{groundstate_index}'].clone()

    return outdict



# -*- coding: utf-8 -*-
'relax': True,               # Not implemented, relax the structure
'relax_mode': 'Fleur',       # Not implemented, how to relax the structure
'relax_para': 'default',     # Not implemented, parameter for the relaxation
'scf_para': 'default',       # Use these parameters for the SCFs
'same_para': True,           # enforce the same parameters
'references': {}             # Dict to provide the elemental references
                             # i.e { 'W': calc, outputnode of SCF workflow or fleurinp,
                             # or structure data or (structure data + Parameter),
                             # 'Be' : ...}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurMaeWorkChain' for the calculation of
    Magnetic Anisotropy Energy via the force theorem.
"""

import copy

#from six.moves import map
from lxml import etree
from ase.dft.kpoints import monkhorst_pack

from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction as cf
from aiida.orm import Code, load_node
from aiida.orm import RemoteData, Dict, KpointsData
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode, get_inputs_fleur
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.data.fleurinp import FleurinpData, get_fleurinp_from_remote_data
from aiida_fleur.common.constants import HTR_TO_EV


class FleurMaeWorkChain(WorkChain):
    """
        This workflow calculates the Magnetic Anisotropy Energy of a structure.
    """

    _workflowversion = '0.3.1'

    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 2 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    _default_wf_para = {
        'sqa_ref': [0.7, 0.7],
        'use_soc_ref': False,
        'sqas_theta': [0.0, 1.57079, 1.57079],
        'sqas_phi': [0.0, 0.0, 1.57079],
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },
        'kmesh_force_theorem': None,
        'use_symmetries_reference': False,
        'soc_off': [],
        'inpxml_changes': [],
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           },
                           namespace='scf')
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('remote', valid_type=RemoteData, required=False)
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('options', valid_type=Dict, required=False)

        spec.outline(cls.start,
                     if_(cls.scf_needed)(
                         cls.converge_scf,
                         cls.force_after_scf,
                     ).else_(
                         cls.force_wo_scf,
                     ), cls.get_results, cls.return_results)

        spec.output('output_mae_wc_para', valid_type=Dict)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(233,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Invalid code node specified, check inpgen and fleur code nodes.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(334, 'ERROR_REFERENCE_CALCULATION_FAILED', message='Reference calculation failed.')
        spec.exit_code(335,
                       'ERROR_REFERENCE_CALCULATION_NOREMOTE',
                       message='Found no reference calculation remote repository.')
        spec.exit_code(336, 'ERROR_FORCE_THEOREM_FAILED', message='Force theorem calculation failed.')

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report(f'INFO: started Magnetic Anisotropy Energy calculation workflow version {self._workflowversion}\n')

        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.t_energydict = []
        self.ctx.mae_thetas = []
        self.ctx.mae_phis = []
        self.ctx.fleuroutuuid = None

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for MAE contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        # switch off SOC on an atom specie
        for atom_label in self.ctx.wf_dict['soc_off']:
            self.ctx.wf_dict['inpxml_changes'].append(('set_species_label', {
                'at_label': atom_label,
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                },
                'create': True
            }))

        # Check if sqas_theta and sqas_phi have the same length
        if len(self.ctx.wf_dict.get('sqas_theta')) != len(self.ctx.wf_dict.get('sqas_phi')):
            error = ('Number of sqas_theta has to be equal to the number of sqas_phi')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # initialize the dictionary using defaults if no options are given
        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        # Check if user gave valid fleur executable
        inputs = self.inputs
        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        # Check if user gave an input setup making any sense
        if 'scf' in inputs:
            self.ctx.scf_needed = True
            if 'remote' in inputs:
                error = 'ERROR: you gave SCF input + remote for the FT'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'fleurinp' in inputs:
                error = 'ERROR: you gave SCF input + fleurinp for the FT'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'remote' not in inputs:
            error = 'ERROR: you gave neither SCF input nor remote for the FT'
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            self.ctx.scf_needed = False

    def scf_needed(self):
        """
        Returns True if SCF WC is needed.
        """
        return self.ctx.scf_needed

    def converge_scf(self):
        """
        Converge charge density with or without SOC.
        Submit a single Fleur calculation to obtain
        a reference for further force theorem calculations.
        """
        inputs = self.get_inputs_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(reference=res)

    def get_inputs_scf(self):
        """
        Initialize inputs for scf workflow:
        wf_param, options, calculation parameters, codes, structure
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []

        soc = self.ctx.wf_dict.get('sqa_ref')
        if not self.ctx.wf_dict.get('use_soc_ref'):
            scf_wf_dict['inpxml_changes'].append(('set_inpchanges', {'change_dict': {'l_soc': False}}))
        else:  # set soc parameters explicitly
            changes_dict = {'theta': soc[0], 'phi': soc[1], 'l_soc': True}
            scf_wf_dict['inpxml_changes'].append(('set_inpchanges', {
                'change_dict': changes_dict,
                'path_spec': {
                    'phi': {
                        'contains': 'soc'
                    },
                    'theta': {
                        'contains': 'soc'
                    }
                }
            }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        if 'structure' in input_scf:
            if 'calc_parameters' in input_scf:
                calc_parameters = input_scf.calc_parameters.get_dict()
            else:
                calc_parameters = {}
            if not self.ctx.wf_dict.get('use_symmetries_reference'):
                # break symmetries, SOC will be removed if not set
                calc_parameters['soc'] = {'theta': soc[0], 'phi': soc[1]}
            input_scf.calc_parameters = Dict(dict=calc_parameters)

        return input_scf

    def change_fleurinp(self):
        """
        This routine sets somethings in the fleurinp file before running a fleur
        calculation.
        """
        if self.ctx.scf_needed:
            try:
                fleurin = self.ctx.reference.outputs.fleurinp
            except NotExistent:
                error = 'Fleurinp generated in the reference calculation is not found.'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED
        else:
            if 'fleurinp' in self.inputs:
                fleurin = self.inputs.fleurinp
            else:
                # In this case only remote is given
                # fleurinp data has to be generated from the remote inp.xml file
                fleurin = get_fleurinp_from_remote_data(self.inputs.remote)

        # copy default changes
        fchanges = self.ctx.wf_dict.get('inpxml_changes', [])

        # add forceTheorem tag into inp.xml
        fchanges.append(('set_complex_tag', {
            'tag_name': 'MAE',
            'create': True,
            'changes': {
                'theta': self.ctx.wf_dict['sqas_theta'],
                'phi': self.ctx.wf_dict['sqas_phi']
            }
        }))
        fchanges.append(('set_inpchanges', {'change_dict': {'itmax': 1, 'l_soc': True}}),)

        if self.ctx.wf_dict['kmesh_force_theorem'] is not None:
            # set k-mesh for the full BZ
            kmesh = KpointsData()
            kmesh.set_kpoints(monkhorst_pack(self.ctx.wf_dict['kmesh_force_theorem']))
            kmesh.store()
            fchanges.append(('set_kpointsdata', {
                'kpointsdata_uuid': kmesh.uuid,
                'switch': True,
                'kpoint_type': 'mesh'
            }))

        # if self.ctx.wf_dict['use_symmetries_reference']:
        #     # remove symmetries from the inp.xml
        #     fchanges.append(('delete_tag', {
        #         'tag_name': 'symOp',
        #         'occurrences': range(1, len(fleurin.inp_dict['cell']['symmetryOperations']))
        #     }))

        if fchanges:  # change inp.xml file
            fleurmode = FleurinpModifier(fleurin)
            try:
                fleurmode.add_task_list(fchanges)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply inpxml_changes'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

            # validate?
            try:
                fleurmode.show(display=False, validate=True)
            except etree.DocumentInvalid:
                error = ('ERROR: input, user wanted inp.xml changes did not validate')
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_FILE
            except ValueError as exc:
                error = ('ERROR: input, user wanted inp.xml changes could not be applied.'
                         f'The following error was raised {exc}')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

            # apply
            out = fleurmode.freeze()
            self.ctx.fleurinp = out
        else:  # otherwise do not change the inp.xml
            self.ctx.fleurinp = fleurin

        return

    def force_after_scf(self):
        """
        Calculate energy of a system for given SQAs
        using the force theorem. Converged reference is stored in self.ctx['xyz'].
        """
        calc = self.ctx.reference

        if not calc.is_finished_ok:
            message = ('The reference SCF calculation was not successful.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        try:
            outpara_node = calc.outputs.output_scf_wc_para
        except NotExistent:
            message = ('The reference SCF calculation failed, no scf output node.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        outpara = outpara_node.get_dict()

        t_e = outpara.get('total_energy', 'failed')
        if not isinstance(t_e, float):
            message = ('Did not manage to extract float total energy from the reference SCF calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder of the reference calculation
        pk_last = 0
        scf_ref_node = load_node(calc.pk)
        for i in scf_ref_node.called:
            if i.node_type == 'process.workflow.workchain.WorkChainNode.':
                if i.process_class is FleurBaseWorkChain:
                    if pk_last < i.pk:
                        pk_last = i.pk
        try:
            remote = load_node(pk_last).outputs.remote_folder
        except AttributeError:
            message = ('Found no remote folder of the reference scf calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_NOREMOTE

        label = 'MAE_force_theorem'
        description = 'This is the force theorem calculation for MAE.'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def force_wo_scf(self):
        """
        Submit FLEUR force theorem calculation using input remote
        """
        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder from the inputs
        remote = self.inputs.remote

        label = 'Force_theorem_calculation'
        description = 'This is a force theorem calculation for all SQA'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def get_results(self):
        """
        Generates results of the workchain.
        """
        t_energydict = []
        mae_thetas = []
        mae_phis = []
        fleur_output_uuid = None

        try:
            calculation = self.ctx.f_t
            if not calculation.is_finished_ok:
                message = f'ERROR: Force theorem Fleur calculation failed somehow it has exit status {calculation.exit_status}'
                self.control_end_wc(message)
                return self.exit_codes.ERROR_FORCE_THEOREM_FAILED
        except AttributeError:
            message = 'ERROR: Something went wrong I do not have a force theorem Fleur calculation'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        try:
            fleurout = calculation.outputs.output_parameters
            fleur_output_uuid = fleurout.uuid
            out_dict = fleurout.dict
            t_energydict = out_dict.mae_force_evsum
            mae_thetas = out_dict.mae_force_theta
            mae_phis = out_dict.mae_force_phi
            e_u = out_dict.mae_force_units

            minenergy = min(t_energydict)

            if e_u in ['Htr', 'htr']:
                t_energydict = [HTR_TO_EV * (x - minenergy) for x in t_energydict]
            else:
                t_energydict = [(x - minenergy) for x in t_energydict]

        except AttributeError as e_message:
            message = f'Did not manage to read evSum or energy units after FT calculation. {e_message}'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        self.ctx.t_energydict = t_energydict
        self.ctx.mae_thetas = mae_thetas
        self.ctx.mae_phis = mae_phis
        self.ctx.fleuroutuuid = fleur_output_uuid

    def return_results(self):
        """
        This function outputs results of the wc
        """

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            # 'initial_structure': self.inputs.structure.uuid,
            'is_it_force_theorem': True,
            'maes': self.ctx.t_energydict,
            'theta': self.ctx.mae_thetas,
            'phi': self.ctx.mae_phis,
            'mae_units': 'eV',
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors
        }

        # ensure provenance of output nodes

        out_dict = {'out': Dict(dict=out)}
        if self.ctx.fleuroutuuid is not None:
            out_dict['last_fleur_out'] = load_node(self.ctx.fleuroutuuid)

        out_nodes = save_mae_output_node(**out_dict)
        out = out_nodes.get('output_mae_wc_para')

        # make wc return out node
        self.out('output_mae_wc_para', out)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.report(errormsg)  # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def save_mae_output_node(**kwargs):
    """
    This is a pseudo cf, to create the right graph structure of AiiDA.
    This calcfunction will create the output node in the database.
    It also connects the output_node to all nodes the information comes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    for key, val in kwargs.items():
        if key == 'out':  # should be always there
            outpara = val
    outdict = {}

    # clone, because we rather produce the same node twice then have a circle in the database for
    outputnode = outpara.clone()
    outputnode.label = 'output_mae_wc_para'
    outputnode.description = ('Contains magnetic anisotropy results and information of an FleurMaeWorkChain run.')

    outdict['output_mae_wc_para'] = outputnode

    return outdict



# -*- coding: utf-8 -*-
{
    'configurations': {'all-Nd-3': [[(1,1,1,1,0,0,0),       #Lists of used configurations
                                     (1,1,1,0,1,0,0),       #for all LDA+U orbitals and spin
                                     ...],
                                    [(0,0,0,0,0,0,0),
                                     (0,0,0,0,0,0,0),
                                     ...]]},
    'total_energy': [
        -38166.542950054,
        -38166.345602746,
        ...
    ],
    'total_energy_units': 'Htr',
    'distance_charge': [
        0.000001,
        0.0000023,
        ...
    ],
    'distance_charge_units': 'me/bohr^3',
    'successful_configs': [0,1,2,3,...],                   #Which configurations successfully converged
    'non_converged_configs': [],                           #Which configurations did not converge
    'failed_configs': [],                                  #Which configurations failed for another reason
    'info': [],
    'warnings': [],
    'errors': [],
    'workflow_name': 'FleurOrbControlWorkChain',
    'workflow_version': '0.1.0'
}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
# pylint: disable=cyclic-import
# ,reimported,unused-import,wrong-import-position
"""
Module with CLI commands for various data types.
"""
import click


@click.group('data')
def cmd_data():
    """Commands to create and inspect data nodes."""


#cmd_root.add_command(cmd_data)

# Import the sub commands to register them with the CLI
from .structure import cmd_structure
from .parameters import cmd_parameter
from .fleurinp import cmd_fleurinp



# -*- coding: utf-8 -*-
'lattice': 'fcc',             # type of the substrate lattice: 'bcc' or 'fcc'
'miller': None,               # miller indices to set up planes forming the p.u.c.
'directions': None,           # miller indices to set up vectors forming the p.u.c.
'host_symbol': 'Pt',          # chemical element of the substrate
'latticeconstant': 4.0,       # initial guess for the substrate lattice constant
'size': (1, 1, 5),            # sets the size of the film unit cell for relax step
'replacements': {1: 'Fe',     # sets the layer number to be replaced by another element
                 -1: 'Fe'},   # NOTE: negative number means that replacement will take place
                              # in the last layer

'decimals': 10,               # set the accuracy of writing atom positions
'z_coordinate_window': 2,     # set how z-coordinates will be rounded before grouping in layers
'pop_last_layers': 1,         # number of bottom layers to be removed before relaxation
'hold_layers': None,          # a list of layer numbers to be held during the relaxation
'AFM_name': 'FM',             # a name of hardcoded AFM structure: FM, AFM_x, AFM_y, AFM_xy
'magnetic_layers': 1,         # the total number of magnetic layers (for symmetric films devide by 2)
'AFM_layer_positions': None,  # a suggestion of initial layer z coordinates, used if AFM structures are known
                              # (relaxXYZ = 'FFF')
'last_layer_factor': 0.85,    # factor by which interlayer distance between two last layers
                              # will be multiplied
'first_layer_factor': 0.0,    # factor by which interlayer distance between two first layers
                              # will be multiplied
'total_number_layers': 4,     # use this total number of layers
'num_relaxed_layers': 2,      # use this number of relaxed interlayer distances



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This is the worklfow 'band' for the Fleur code, which calculates a
electron bandstructure.
"""
# TODO alow certain kpoint path, or kpoint node, so far auto
import copy
from lxml import etree
from ase.dft.kpoints import bandpath
import numpy as np

from aiida.orm import Code, Dict, RemoteData, KpointsData
from aiida.orm import load_node, CalcJobNode, FolderData, BandsData, XyData
from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction as cf
from aiida.common.exceptions import NotExistent
from aiida.common import AttributeDict
from aiida.tools.data.array.kpoints import get_explicit_kpoints_path

from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.data.fleurinp import FleurinpData, get_fleurinp_from_remote_data


class FleurBandDosWorkChain(WorkChain):
    '''
    This workflow calculated a bandstructure from a Fleur calculation

    :Params: a Fleurcalculation node
    :returns: Success, last result node, list with convergence behavior
    '''
    # wf_parameters: {  'tria', 'nkpts', 'sigma', 'emin', 'emax'}
    # defaults : tria = True, nkpts = 800, sigma=0.005, emin= , emax =

    _workflowversion = '0.5.2'

    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }
    _default_wf_para = {
        'kpath': 'auto',  #seek (aiida), fleur (only Max4) or string to pass to ase
        'mode': 'band',
        'klistname': None,
        'kpoints_number': None,
        'kpoints_distance': None,
        'kpoints_explicit': None,  #dictionary containing a list of kpoints, weights
        #and additional arguments to pass to set_kpointlist
        'sigma': 0.005,
        'emin': -0.50,
        'emax': 0.90,
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },
        'inpxml_changes': [],
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           },
                           namespace='scf')
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('remote', valid_type=RemoteData, required=False)
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('kpoints', valid_type=KpointsData, required=False)
        spec.input('options', valid_type=Dict, required=False)

        spec.outline(cls.start,
                     if_(cls.scf_needed)(
                         cls.converge_scf,
                         cls.banddos_after_scf,
                     ).else_(
                         cls.banddos_wo_scf,
                     ), cls.return_results)

        spec.output('output_banddos_wc_para', valid_type=Dict)
        spec.output('last_calc_retrieved', valid_type=FolderData)
        spec.output('output_banddos_wc_bands', valid_type=BandsData, required=False)
        spec.output('output_banddos_wc_dos', valid_type=XyData, required=False)

        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(233,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Invalid code node specified, check inpgen and fleur code nodes.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(334, 'ERROR_SCF_CALCULATION_FAILED', message='SCF calculation failed.')
        spec.exit_code(335, 'ERROR_SCF_CALCULATION_NOREMOTE', message='Found no SCF calculation remote repository.')

    def start(self):
        '''
        check parameters, what condictions? complete?
        check input nodes
        '''
        ### input check ### ? or done automaticly, how optional?
        # check if fleuinp corresponds to fleur_calc
        self.report(f'started bandsdos workflow version {self._workflowversion}')
        #print("Workchain node identifiers: ")#'{}'
        #"".format(ProcessRegistry().current_calc_node))

        self.ctx.scf_needed = False
        self.ctx.banddos_calc = None
        self.ctx.fleurinp_banddos = None
        self.ctx.scf = None
        self.ctx.successful = False
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []

        inputs = self.inputs

        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in inputs:
            wf_dict = inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        extra_keys = []
        for key in self.ctx.wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for Banddos contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        defaultoptions = self._default_options
        if 'options' in inputs:
            options = inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = 'The code you provided for FLEUR does not use the plugin fleur.fleur'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        if 'scf' in inputs:
            self.ctx.scf_needed = True
            if 'remote' in inputs:
                error = 'ERROR: you gave SCF input + remote for the BandDOS calculation'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'fleurinp' in inputs:
                error = 'ERROR: you gave SCF input + fleurinp for the BandDOS calculation'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'remote' not in inputs:
            error = 'ERROR: you gave neither SCF input nor remote'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            self.ctx.scf_needed = False

        if wf_dict['mode'] == 'dos' and wf_dict['kpath'] not in ('auto', 'skip'):
            error = 'ERROR: you specified the DOS mode but provided a non default kpath argument'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        if wf_dict['kpoints_number'] is not None and wf_dict['kpoints_distance'] is not None:
            error = 'ERROR: Only provide either the distance or number for the kpoints'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

    def change_fleurinp(self):
        """
        create a new fleurinp from the old with certain parameters
        """
        # TODO allow change of kpoint mesh?, tria?
        wf_dict = self.ctx.wf_dict

        if self.ctx.scf_needed:
            try:
                fleurin = self.ctx.scf.outputs.fleurinp
            except NotExistent:
                error = 'Fleurinp generated in the SCF calculation is not found.'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_SCF_CALCULATION_FAILED
        else:
            if 'fleurinp' not in self.inputs:
                fleurin = get_fleurinp_from_remote_data(self.inputs.remote)
            else:
                fleurin = self.inputs.fleurinp

        # how can the user say he want to use the given kpoint mesh, ZZ nkpts : False/0
        fleurmode = FleurinpModifier(fleurin)

        fchanges = wf_dict.get('inpxml_changes', [])
        # apply further user dependend changes
        if fchanges:
            try:
                fleurmode.add_task_list(fchanges)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply inpxml_changes'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        kpath = wf_dict['kpath']
        explicit = wf_dict['kpoints_explicit']
        distance = wf_dict['kpoints_distance']
        nkpts = wf_dict['kpoints_number']
        listname = wf_dict['klistname']

        if explicit is not None:
            try:
                fleurmode.set_kpointlist(**explicit)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply kpoints_explicit'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        if listname is None and wf_dict['mode'] == 'band':
            listname = 'path-2'

        if nkpts is None and distance is None:
            nkpts = 500

        if 'kpoints' in self.inputs:
            kpoint_type = 'path' if wf_dict['mode'] == 'band' else 'mesh'
            fleurmode.set_kpointsdata(self.inputs.kpoints, switch=True, kpoint_type=kpoint_type)
        elif kpath == 'auto':
            if fleurin.inp_version >= '0.32' and listname is not None:
                fleurmode.switch_kpointset(listname)
        elif isinstance(kpath, dict):
            if fleurin.inp_version < '0.32':
                if distance is not None:
                    raise ValueError('set_kpath only supports specifying the number of points for the kpoints')
                fleurmode.set_kpath(kpath, nkpts)
            else:
                raise ValueError('set_kpath is only supported for inputs up to Max4')
        elif kpath == 'seek':
            #Use aiida functionality
            struc = fleurin.get_structuredata_ncf()

            if distance is not None:
                output = get_explicit_kpoints_path(struc, reference_distance=distance)
            else:
                output = get_explicit_kpoints_path(struc)
            primitive_struc = output['primitive_structure']

            #check if primitive_structure and input structure are identical:
            maxdiff_cell = sum(abs(np.array(primitive_struc.cell) - np.array(struc.cell))).max()

            if maxdiff_cell > 3e-9:
                self.report(f'Error in cell : {maxdiff_cell}')
                self.report(
                    'WARNING: The structure data from the fleurinp is not the primitive structure type, which is mandatory in some cases'
                )

            output['explicit_kpoints'].store()

            fleurmode.set_kpointsdata(output['explicit_kpoints'], switch=True)

        elif kpath == 'skip':
            pass
        else:
            #Use ase
            struc = fleurin.get_structuredata_ncf()

            path = bandpath(kpath, cell=struc.cell, npoints=nkpts, density=distance)

            special_points = path.special_points

            labels = []
            for label, special_kpoint in special_points.items():
                for index, kpoint in enumerate(path.kpts):
                    if sum(abs(np.array(special_kpoint) - np.array(kpoint))).max() < 1e-12:
                        labels.append((index, label))
            labels = sorted(labels, key=lambda x: x[0])

            kpts = KpointsData()
            kpts.set_cell(struc.cell)
            kpts.pbc = struc.pbc
            weights = np.ones(len(path.kpts)) / len(path.kpts)
            kpts.set_kpoints(kpoints=path.kpts, cartesian=False, weights=weights, labels=labels)

            kpts.store()
            fleurmode.set_kpointsdata(kpts, switch=True)

        sigma = wf_dict['sigma']
        emin = wf_dict['emin']
        emax = wf_dict['emax']

        if fleurin.inp_version < '0.32' and wf_dict['mode'] == 'dos':
            fleurmode.set_inpchanges({'ndir': -1})

        if wf_dict['mode'] == 'dos':
            change_dict = {'dos': True, 'minEnergy': emin, 'maxEnergy': emax, 'sigma': sigma}
        else:
            change_dict = {'band': True, 'minEnergy': emin, 'maxEnergy': emax, 'sigma': sigma}
        fleurmode.set_inpchanges(change_dict)

        try:
            fleurmode.show(display=False, validate=True)
        except etree.DocumentInvalid:
            error = ('ERROR: input, user wanted inp.xml changes did not validate')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_FILE
        except ValueError as exc:
            error = ('ERROR: input, user wanted inp.xml changes could not be applied.'
                     f'The following error was raised {exc}')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        fleurinp_new = fleurmode.freeze()
        self.ctx.fleurinp_banddos = fleurinp_new

    def scf_needed(self):
        """
        Returns True if SCF WC is needed.
        """
        return self.ctx.scf_needed

    def converge_scf(self):
        """
        Converge charge density.
        """
        inputs = self.get_inputs_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(scf=res)

    def banddos_after_scf(self):
        """
        This method submits the BandDOS calculation after the initial SCF calculation
        """
        calc = self.ctx.scf

        if not calc.is_finished_ok:
            message = ('The SCF calculation was not successful.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_SCF_CALCULATION_FAILED

        try:
            outpara_node = calc.outputs.output_scf_wc_para
        except NotExistent:
            message = ('The SCF calculation failed, no scf output node.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_SCF_CALCULATION_FAILED

        outpara = outpara_node.get_dict()

        if 'total_energy' not in outpara:
            message = ('Did not manage to extract float total energy from the SCF calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_SCF_CALCULATION_FAILED

        self.report('INFO: run BandDOS calculation')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp_banddos
        if fleurin is None:
            error = ('ERROR: Creating BandDOS Fleurinp failed for an unknown reason')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder of the reference calculation
        pk_last = 0
        scf_ref_node = load_node(calc.pk)
        for i in scf_ref_node.called:
            if i.node_type == 'process.workflow.workchain.WorkChainNode.':
                if i.process_class is FleurBaseWorkChain:
                    if pk_last < i.pk:
                        pk_last = i.pk
        try:
            remote = load_node(pk_last).outputs.remote_folder
        except AttributeError:
            message = ('Found no remote folder of the reference scf calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_SCF_CALCULATION_NOREMOTE

        label = 'bansddos_calculation'
        description = 'Bandstructure or DOS is calculated for the given structure'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(banddos_calc=future)

    def banddos_wo_scf(self):
        """
        This method submits the BandDOS calculation without a previous SCF calculation
        """
        self.report('INFO: run BandDOS calculation')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp_banddos
        if fleurin is None:
            error = ('ERROR: Creating BandDOS Fleurinp failed for an unknown reason')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder from the inputs
        remote = self.inputs.remote

        label = 'bansddos_calculation'
        description = 'Bandstructure or DOS is calculated for the given structure'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(banddos_calc=future)

    def get_inputs_scf(self):
        """
        Initialize inputs for scf workflow:
        wf_param, options, calculation parameters, codes, structure
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        return input_scf

    def return_results(self):
        '''
        return the results of the calculations
        '''
        # TODO more here
        self.report('BandDOS workflow Done')

        from aiida_fleur.tools.common_fleur_wf import find_last_submitted_calcjob
        if self.ctx.banddos_calc:
            self.report(f'A bandstructure/DOS was calculated and is found under pk={self.ctx.banddos_calc.pk}, '
                        f'calculation {self.ctx.banddos_calc}')
            try:
                last_calc_uuid = find_last_submitted_calcjob(self.ctx.banddos_calc)
            except NotExistent:
                last_calc_uuid = None
        else:
            last_calc_uuid = None

        try:  # if something failed, we still might be able to retrieve something
            last_calc_out = self.ctx.banddos_calc.outputs.output_parameters
            retrieved = self.ctx.banddos_calc.outputs.retrieved
            if 'fleurinpdata' in self.ctx.banddos_calc.inputs:
                fleurinp = self.ctx.banddos_calc.inputs.fleurinpdata
            else:
                fleurinp = get_fleurinp_from_remote_data(self.ctx.banddos_calc.inputs.parent_folder)
            last_calc_out_dict = last_calc_out.get_dict()
        except (NotExistent, AttributeError):
            last_calc_out = None
            last_calc_out_dict = {}
            retrieved = None
            fleurinp = None

        #check if band file exists: if not succesful = False
        #TODO be careful with general bands.X
        bandfiles = ['bands.1', 'bands.2', 'banddos.hdf', 'Local.1', 'Local.2']

        bandfile_res = []
        if retrieved:
            bandfile_res = retrieved.list_object_names()

        for name in bandfiles:
            if name in bandfile_res:
                self.ctx.successful = True
        if not self.ctx.successful:
            self.report('!NO bandstructure/DOS file was found, something went wrong!')

        # # get efermi from last calculation
        scf_results = None
        efermi_scf = 0
        bandgap_scf = 0
        if 'remote' in self.inputs:
            scf_results = self.inputs.remote.creator.res
        elif 'scf' in self.inputs:
            if self.ctx.scf and self.ctx.scf.is_finished_ok:
                scf_results = self.ctx.scf.outputs.last_calc.output_parameters.dict

        if scf_results is not None:
            efermi_scf = scf_results.fermi_energy
            bandgap_scf = scf_results.bandgap

        efermi_band = last_calc_out_dict.get('fermi_energy', None)
        bandgap_band = last_calc_out_dict.get('bandgap', None)

        diff_efermi = None
        if efermi_band is not None:
            diff_efermi = efermi_scf - efermi_band

        diff_bandgap = None
        if bandgap_band is not None:
            diff_bandgap = bandgap_scf - bandgap_band

        outputnode_dict = {}

        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['Warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['last_calc_uuid'] = last_calc_uuid
        outputnode_dict['mode'] = self.ctx.wf_dict.get('mode')
        outputnode_dict['fermi_energy_band'] = efermi_band
        outputnode_dict['bandgap_band'] = bandgap_band
        outputnode_dict['fermi_energy_scf'] = efermi_scf
        outputnode_dict['bandgap_scf'] = bandgap_scf
        outputnode_dict['diff_efermi'] = diff_efermi
        outputnode_dict['diff_bandgap'] = diff_bandgap
        outputnode_dict['bandgap_units'] = 'eV'
        outputnode_dict['fermi_energy_units'] = 'Htr'

        outputnode_t = Dict(dict=outputnode_dict)
        if last_calc_out:
            outdict = create_band_result_node(outpara=outputnode_t,
                                              last_calc_out=last_calc_out,
                                              last_calc_retrieved=retrieved)

            if self.ctx.wf_dict.get('mode') == 'band' and fleurinp is not None and retrieved is not None:
                bands = create_aiida_bands_data(fleurinp=fleurinp, retrieved=retrieved)
                if isinstance(bands, BandsData):
                    outdict['output_banddos_wc_bands'] = bands
            elif self.ctx.wf_dict.get('mode') == 'dos' and retrieved is not None:
                dos = create_aiida_dos_data(retrieved=retrieved)
                if isinstance(dos, XyData):
                    outdict['output_banddos_wc_dos'] = dos

        else:
            outdict = create_band_result_node(outpara=outputnode_t)

        if retrieved:
            outdict['last_calc_retrieved'] = retrieved

        #TODO parse Bandstructure
        for link_name, node in outdict.items():
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.report(errormsg)  # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_band_result_node(**kwargs):
    """
    This is a pseudo wf, to create the right graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    for key, val in kwargs.items():
        if key == 'outpara':  # should be always there
            outpara = val
    outdict = {}
    outputnode = outpara.clone()
    outputnode.label = 'output_banddos_wc_para'
    outputnode.description = ('Contains band calculation results')

    outdict['output_banddos_wc_para'] = outputnode

    return outdict


@cf
def create_aiida_bands_data(fleurinp, retrieved):
    """
    Creates :py:class:`aiida.orm.BandsData` object containing the kpoints and eigenvalues
    from the `banddos.hdf` file of the calculation

    :param fleurinp: :py:class:`~aiida_fleur.data.fleurinp.FleurinpData` for the calculation
    :param retrieved: :py:class:`aiida.orm.FolderData` for the bandstructure calculation

    :returns: :py:class:`aiida.orm.BandsData` for the bandstructure calculation

    :raises: ExitCode 300, banddos.hdf file is missing
    :raises: ExitCode 310, banddos.hdf reading failed
    :raises: ExitCode 320, reading kpointsdata from Fleurinp failed
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader, HDF5TransformationError
    from masci_tools.io.parsers.hdf5.recipes import FleurSimpleBands  #no projections only eigenvalues for now
    from aiida.engine import ExitCode

    try:
        kpoints = fleurinp.get_kpointsdata_ncf(only_used=True)
    except ValueError as exc:
        return ExitCode(320, message=f'Retrieving kpoints data from fleurinp failed with: {exc}')

    if 'banddos.hdf' in retrieved.list_object_names():
        try:
            with retrieved.open('banddos.hdf', 'rb') as f:
                with HDF5Reader(f) as reader:
                    data, attributes = reader.read(recipe=FleurSimpleBands)
        except (HDF5TransformationError, ValueError) as exc:
            return ExitCode(310, message=f'banddos.hdf reading failed with: {exc}')
    else:
        return ExitCode(300, message='banddos.hdf file not in the retrieved files')

    bands = BandsData()
    bands.set_kpointsdata(kpoints)

    nkpts, nbands = attributes['nkpts'], attributes['nbands']
    eigenvalues = data['eigenvalues_up'].reshape((nkpts, nbands))
    if 'eigenvalues_down' in data:
        eigenvalues_dn = data['eigenvalues_down'].reshape((nkpts, nbands))
        eigenvalues = [eigenvalues, eigenvalues_dn]

    bands.set_bands(eigenvalues, units='eV')

    bands.label = 'output_banddos_wc_bands'
    bands.description = ('Contains BandsData for the bandstructure calculation')

    return bands


@cf
def create_aiida_dos_data(retrieved):
    """
    Creates :py:class:`aiida.orm.XyData` object containing the standard DOS components
    from the `banddos.hdf` file of the calculation

    :param retrieved: :py:class:`aiida.orm.FolderData` for the DOS calculation

    :returns: :py:class:`aiida.orm.XyData` containing all standard DOS components

    :raises: ExitCode 300, banddos.hdf file is missing
    :raises: ExitCode 310, banddos.hdf reading failed
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader, HDF5TransformationError
    from masci_tools.io.parsers.hdf5.recipes import FleurDOS  #only standard DOS for now
    from aiida.engine import ExitCode

    if 'banddos.hdf' in retrieved.list_object_names():
        try:
            with retrieved.open('banddos.hdf', 'rb') as f:
                with HDF5Reader(f) as reader:
                    data, attributes = reader.read(recipe=FleurDOS)
        except (HDF5TransformationError, ValueError) as exc:
            return ExitCode(310, message=f'banddos.hdf reading failed with: {exc}')
    else:
        return ExitCode(300, message='banddos.hdf file not in the retrieved files')

    dos = XyData()
    dos.set_x(data['energy_grid'], 'energy', x_units='eV')

    names = [key for key in data if key != 'energy_grid']
    arrays = [entry for key, entry in data.items() if key != 'energy_grid']
    units = ['1/eV'] * len(names)
    dos.set_y(arrays, names, y_units=units)

    dos.label = 'output_banddos_wc_dos'
    dos.description = (
        'Contains XyData for the density of states calculation with total, interstitial, atom and orbital weights')

    return dos



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Input plug-in for the FLEUR input generator 'inpgen'.
The input generator for the Fleur code is a preprocessor
and should be run locally (with the direct scheduler) or inline,
because it does not take many resources.
"""
from aiida.engine import CalcJob
from aiida.common.exceptions import InputValidationError
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.orm import StructureData, Dict

from aiida_fleur.data.fleurinp import FleurinpData
import io


class FleurinputgenCalculation(CalcJob):
    """
    JobCalculationClass for the inpgen, which is a preprocessor for a FLEUR calculation.
    For more information about produced files and the FLEUR-code family, go to http://www.flapw.de/.
    """

    __version__ = '1.2.2'

    # Default input and output files
    _INPUT_FILE = 'aiida.in'  # will be shown with inputcat
    _OUTPUT_FILE = 'out'  # 'shell.out' #will be shown with outputcat

    # created file names, some needed for Fleur calc
    _INPXML_FILE_NAME = 'inp.xml'
    _INPUT_FILE_NAME = 'aiida.in'
    _SHELLOUT_FILE_NAME = 'shell.out'
    _OUTPUT_FILE_NAME = 'out'
    _ERROR_FILE_NAME = 'out.error'
    _STRUCT_FILE_NAME = 'struct.xsf'
    _JUDFT_WARN_ONLY_INFO_FILE_NAME = 'JUDFT_WARN_ONLY'

    _settings_keys = [
        'additional_retrieve_list', 'remove_from_retrieve_list', 'cmdline', 'significant_figures_cell',
        'significant_figures_positions', 'profile'
    ]
    # TODO switch all these to init_internal_params?
    _OUTPUT_SUBFOLDER = './fleur_inp_out/'
    _PREFIX = 'aiida'

    # Additional files that should always be retrieved for the specific plugin
    _internal_retrieve_list = []
    _automatic_namelists = {}

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('metadata.options.input_filename', valid_type=str, default=cls._INPUT_FILE)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._INPXML_FILE_NAME)
        spec.input('structure', valid_type=StructureData, help='Choose the input structure to use')
        spec.input('parameters',
                   valid_type=Dict,
                   required=False,
                   help='Use a node that specifies the input parameters '
                   'for the namelists')
        spec.input('settings',
                   valid_type=Dict,
                   required=False,
                   help='This parameter data node is used to specify for some '
                   'advanced features how the plugin behaves. You can add files'
                   'the retrieve list, or add command line switches, '
                   'for all available features here check the documentation.')

        # parser
        spec.input('metadata.options.parser_name', valid_type=str, default='fleur.fleurinpgenparser')

        # declaration of outputs of the calclation
        #TODO: Should this be renamed? FleurinpData inputs/outputs are called fleurinp everywhere
        #else
        spec.output('fleurinpData', valid_type=FleurinpData, required=True)

        # exit codes
        # spec.exit_code(251, 'ERROR_WRONG_INPUT_PARAMS',
        #                message='Input parameters for inpgen contain unknown keys.')
        # spec.exit_code(253, 'ERROR_ATOM_POSITION_NEEDED',
        #                message='Fleur lattice needs atom positions as input.')
        # spec.exit_code(254, 'ERROR_INPUT_PARAMS_LEFTOVER',
        #                message='Excessive input parameters were specified.')
        spec.exit_code(300, 'ERROR_NO_RETRIEVED_FOLDER', message='No retrieved folder found.')
        spec.exit_code(301, 'ERROR_OPENING_OUTPUTS', message='One of the output files can not be opened.')
        spec.exit_code(306, 'ERROR_NO_INPXML', message='XML input file was not found.')
        spec.exit_code(307, 'ERROR_MISSING_RETRIEVED_FILES', message='Some required files were not retrieved.')
        spec.exit_code(308,
                       'ERROR_FLEURINPDATA_INPUT_NOT_VALID',
                       message=('During parsing: FleurinpData could not be initialized, see log. '))
        spec.exit_code(309, 'ERROR_FLEURINPDATA_NOT_VALID', message='During parsing: FleurinpData failed validation.')
        spec.exit_code(310,
                       'ERROR_UNKNOWN_PROFILE',
                       message='The profile {profile} is not known to the used inpgen code')

    def prepare_for_submission(self, folder):
        """
        This is the routine to be called when you want to create
        the input files for the inpgen with the plug-in.

        :param folder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        """

        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []

        # first check existence of structure and if 1D, 2D, 3D
        structure = self.inputs.structure

        # check existence of parameters (optional)
        if 'parameters' in self.inputs:
            parameters = self.inputs.parameters
        else:
            parameters = None

        if parameters is None:
            # use default
            parameters_dict = {}
        else:
            parameters_dict = _lowercase_dict(parameters.get_dict(), dict_name='parameters')

        code = self.inputs.code

        # check existence of settings (optional)
        if 'settings' in self.inputs:
            settings = self.inputs.settings
        else:
            settings = None

        if settings is None:
            settings_dict = {}
        else:
            settings_dict = settings.get_dict()

        # check for for allowed keys, ignore unknown keys but warn.
        for key in settings_dict.keys():
            if key not in self._settings_keys:
                # TODO warning
                self.logger.info('settings dict key %s for Fleur calculation'
                                 'not recognized, only %s are allowed.', key, str(self._settings_keys))

        #Check that if a inpgen profile is given no additional parameters are provided
        #since this will overwrite the effects of the inpgen profile (For now we just issue a warning)
        if 'profile' in settings_dict:
            lapw_parameters_given = any('atom' in key for key in parameters_dict)
            comp = parameters_dict.get('comp', {})
            lapw_parameters_given = lapw_parameters_given or \
                                    'kmax' in comp or \
                                    'gmax' in comp or \
                                    'gmaxxc' in comp
            if lapw_parameters_given:
                self.logger.warning('Inpgen profile specified but atom/LAPW basis specific '
                                    'parameters are provided. These will conflict/override each other')

        #######################################
        #### WRITE ALL CARDS IN INPUT FILE ####

        with folder.open(self._INPUT_FILE_NAME, 'w') as input_file:
            write_inpgen_file_aiida_struct(structure, input_file, input_params=parameters_dict, settings=settings_dict)

        # create a JUDFT_WARN_ONLY file in the calculation folder
        with io.StringIO('/n') as handle:
            warn_only_filename = self._JUDFT_WARN_ONLY_INFO_FILE_NAME
            folder.create_file_from_filelike(handle, filename=warn_only_filename, mode='w')

        calcinfo = CalcInfo()

        calcinfo.uuid = self.uuid

        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = remote_symlink_list

        # Retrieve per default only out file and inp.xml file?
        retrieve_list = []
        retrieve_list.append(self._INPXML_FILE_NAME)
        retrieve_list.append(self._OUTPUT_FILE_NAME)
        retrieve_list.append(self._SHELLOUT_FILE_NAME)
        retrieve_list.append(self._ERROR_FILE_NAME)
        retrieve_list.append(self._STRUCT_FILE_NAME)
        retrieve_list.append(self._INPUT_FILE_NAME)

        # user specific retrieve
        add_retrieve = settings_dict.get('additional_retrieve_list', [])
        for file1 in add_retrieve:
            retrieve_list.append(file1)

        remove_retrieve = settings_dict.get('remove_from_retrieve_list', [])
        for file1 in remove_retrieve:
            if file1 in retrieve_list:
                retrieve_list.remove(file1)

        calcinfo.retrieve_list = []
        for file1 in retrieve_list:
            calcinfo.retrieve_list.append(file1)

        codeinfo = CodeInfo()
        # , "-electronConfig"] # TODO? let the user decide -electronconfig?

        # We support different inpgen and fleur version via reading the version from the code node extras
        code_extras = code.extras
        code_version = code_extras.get('version', 32)
        if int(code_version) < 32:
            # run old inpgen
            cmdline_params = ['-explicit']
            codeinfo.stdin_name = self._INPUT_FILE_NAME
        else:
            cmdline_params = ['-explicit', '-inc', '+all', '-f', f'{self._INPUT_FILE_NAME}']

        if 'profile' in settings_dict:
            cmdline_params.extend(['-profile', settings_dict['profile']])

        # user specific commandline_options
        for command in settings_dict.get('cmdline', []):
            cmdline_params.append(command)
        codeinfo.cmdline_params = (list(cmdline_params))

        codeinfo.code_uuid = code.uuid
        codeinfo.stdout_name = self._SHELLOUT_FILE_NAME  # shell output will be piped in file
        codeinfo.stderr_name = self._ERROR_FILE_NAME  # std error too
        calcinfo.codes_info = [codeinfo]

        return calcinfo


def _lowercase_dict(dic, dict_name):
    """
    Converts every entry in a dictionary to lowercase

    :param dic: parameters dictionary
    :param dict_name: dictionary name
    """
    from collections import Counter

    if isinstance(dic, dict):
        new_dict = {str(k).lower(): val for k, val in dic.items()}
        if len(new_dict) != len(dic):
            num_items = Counter(str(k).lower() for k in dic.keys())
            double_keys = ','.join([k for k, val in num_items if val > 1])
            raise InputValidationError("Inside the dictionary '{}' there are the following keys that "
                                       'are repeated more than once when compared case-insensitively:'
                                       '{}.This is not allowed.'.format(dict_name, double_keys))
        return new_dict
    else:
        raise TypeError('_lowercase_dict accepts only dictionaries as argument')


def write_inpgen_file_aiida_struct(structure, file, input_params=None, settings=None):
    """Wraps around masci_tools write inpgen_file, unpacks aiida structure"""
    from masci_tools.io.fleur_inpgen import write_inpgen_file

    atoms_dict_list = []
    kind_list = []

    for kind in structure.kinds:
        kind_list.append(kind.get_raw())

    for site in structure.sites:
        atoms_dict_list.append(site.get_raw())

    if settings is None:
        settings = {}

    write_settings = {}
    if 'significant_figures_cell' in settings:
        write_settings['significant_figures_cell'] = settings.get('significant_figures_cell')
    if 'significant_figures_positions' in settings:
        write_settings['significant_figures_positions'] = settings.get('significant_figures_positions')

    report = write_inpgen_file(structure.cell,
                               atoms_dict_list,
                               kind_list,
                               file=file,
                               pbc=structure.pbc,
                               input_params=input_params,
                               **write_settings)

    return report



# -*- coding: utf-8 -*-
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida.orm import Dict, load_node

fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)
structure = load_node(STRUCTURE_PK)

wf_para = Dict(dict={'fleur_runmax': 3,
                     'density_converged': 0.001,
                     'mode': 'density',
                     'itmax_per_run': 30,
                     'add_comp_para': {
                                        'only_even_MPI': False,
                                        'max_queue_nodes': 20,
                                        'max_queue_wallclock_sec': 86400
                                        }})

options = Dict(dict={'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 2},
                     'withmpi': True,
                     'max_wallclock_seconds': 600})

calc_parameters = Dict(dict={'kpt': {'div1': 2,
                                     'div2': 2,
                                     'div3': 2
                                     }})

SCF_workchain = submit(FleurScfWorkChain,
                       fleur=fleur_code,
                       inpgen=inpgen_code,
                       calc_parameters=calc_parameters,
                       structure=structure,
                       wf_parameters=wf_para,
                       options=options)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurDMIWorkChain' for the calculation of
    DMI energy dispersion.
"""
import copy
import numpy as np

from lxml import etree
from ase.dft.kpoints import monkhorst_pack

from aiida.engine import WorkChain, ToContext, if_
from aiida.engine import calcfunction as cf
from aiida.orm import Code, load_node
from aiida.orm import RemoteData, Dict, KpointsData
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain
from aiida_fleur.common.constants import HTR_TO_EV
from aiida_fleur.data.fleurinp import FleurinpData, get_fleurinp_from_remote_data


class FleurDMIWorkChain(WorkChain):
    """
    This workflow calculates DMI energy dispersion of a structure.
    """

    _workflowversion = '0.2.1'

    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 2 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    _default_wf_para = {
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },
        'beta': {
            'all': 1.57079
        },
        'sqas_theta': [0.0, 1.57079, 1.57079],
        'sqas_phi': [0.0, 0.0, 1.57079],
        'soc_off': [],
        # 'prop_dir': [1.0, 0.0, 0.0],
        'q_vectors': [[0.0, 0.0, 0.0], [0.125, 0.0, 0.0], [0.250, 0.0, 0.0], [0.375, 0.0, 0.0]],
        'ref_qss': [0.0, 0.0, 0.0],
        'kmesh_force_theorem': None,
        'inpxml_changes': []
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           },
                           namespace='scf')
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('remote', valid_type=RemoteData, required=False)
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('options', valid_type=Dict, required=False)

        spec.outline(cls.start,
                     if_(cls.scf_needed)(
                         cls.converge_scf,
                         cls.force_after_scf,
                     ).else_(
                         cls.force_wo_scf,
                     ), cls.get_results, cls.return_results)

        spec.output('out', valid_type=Dict)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(233,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Invalid code node specified, check inpgen and fleur code nodes.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(334, 'ERROR_REFERENCE_CALCULATION_FAILED', message='Reference calculation failed.')
        spec.exit_code(335,
                       'ERROR_REFERENCE_CALCULATION_NOREMOTE',
                       message='Found no reference calculation remote repository.')
        spec.exit_code(336, 'ERROR_FORCE_THEOREM_FAILED', message='Force theorem calculation failed.')

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report(f'INFO: started DMI calculation workflow version {self._workflowversion}\n')
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.qs = []
        self.ctx.mae_thetas = []
        self.ctx.mae_phis = []
        self.ctx.num_ang = 1
        self.ctx.h_so = []
        self.ctx.q_vectors = []

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for DMI contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        if wf_dict['ref_qss'] != wf_dict['q_vectors'][0]:
            error = ('The first q_vector of the forceTheorem step has to be equal to'
                     'the q vector of the reference calculation.')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # Check if sqas_theta and sqas_phi have the same length
        if len(self.ctx.wf_dict.get('sqas_theta')) != len(self.ctx.wf_dict.get('sqas_phi')):
            error = ('Number of sqas_theta has to be equal to the number of sqas_phi')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # initialize the dictionary using defaults if no options are given
        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        # Check if user gave valid fleur executable
        inputs = self.inputs
        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        # Check if user gave an input setup making any sense
        if 'scf' in inputs:
            self.ctx.scf_needed = True
            if 'remote' in inputs:
                error = 'ERROR: you gave SCF input + remote for the FT'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'fleurinp' in inputs:
                error = 'ERROR: you gave SCF input + fleurinp for the FT'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        elif 'remote' not in inputs:
            error = 'ERROR: you gave neither SCF input nor remote for the FT'
            self.control_end_wc(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            self.ctx.scf_needed = False

    def scf_needed(self):
        """
        Returns True if SCF WC is needed.
        """
        return self.ctx.scf_needed

    def converge_scf(self):
        """
        Converge charge density for collinear case which is a reference for futher
        spin spiral calculations.
        """
        inputs = self.get_inputs_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(reference=res)

    def get_inputs_scf(self):
        """
        Initialize inputs for the scf cycle
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []

        # set up q vector for the reference calculation
        list_ref_qss = self.ctx.wf_dict['ref_qss']
        if [x for x in list_ref_qss if x != 0]:
            changes_dict = {
                'qss': self.ctx.wf_dict['ref_qss'],
                'l_noco': True,
                'ctail': False,
                'l_ss': True,
                'l_soc': False
            }
        else:
            changes_dict = {'qss': ' 0.0 0.0 0.0 ', 'l_noco': False, 'ctail': True, 'l_ss': False, 'l_soc': False}

        scf_wf_dict['inpxml_changes'].append(('set_inpchanges', {'change_dict': changes_dict}))

        # change beta parameter
        for key, val in self.ctx.wf_dict.get('beta', {}).items():
            scf_wf_dict['inpxml_changes'].append(('set_atomgroup_label', {
                'attributedict': {
                    'nocoParams': {
                        'beta': val
                    }
                },
                'atom_label': key
            }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        if 'structure' in input_scf:  # add info about spin spiral propagation
            if 'calc_parameters' in input_scf:
                calc_parameters = input_scf.calc_parameters.get_dict()
            else:
                calc_parameters = {}
            sum_vec = np.array([np.pi / 4.0, np.e / 3.0, np.euler_gamma])
            calc_parameters['qss'] = {'x': sum_vec[0], 'y': sum_vec[1], 'z': sum_vec[2]}
            calc_parameters['soc'] = {'theta': 0.7, 'phi': 0.7}
            input_scf.calc_parameters = Dict(dict=calc_parameters)
        return input_scf

    def change_fleurinp(self):
        """
        This routine sets somethings in the fleurinp file before running a fleur
        calculation.
        """
        self.report('INFO: run change_fleurinp')

        if self.ctx.scf_needed:
            try:
                fleurin = self.ctx.reference.outputs.fleurinp
            except NotExistent:
                error = 'Fleurinp generated in the reference calculation is not found.'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED
        else:
            if 'fleurinp' in self.inputs:
                fleurin = self.inputs.fleurinp
            else:
                # In this case only remote is given
                # fleurinp data has to be generated from the remote inp.xml file
                fleurin = get_fleurinp_from_remote_data(self.inputs.remote)

        # copy inpchanges from wf parameters
        fchanges = self.ctx.wf_dict.get('inpxml_changes', [])
        # create forceTheorem tags
        fchanges.append(('set_complex_tag', {
            'tag_name': 'DMI',
            'create': True,
            'changes': {
                'qVectors': {
                    'q': self.ctx.wf_dict['q_vectors']
                },
                'theta': self.ctx.wf_dict['sqas_theta'],
                'phi': self.ctx.wf_dict['sqas_phi']
            }
        }))

        changes_dict = {
            'itmax': 1,
            'l_noco': True,
            'ctail': False,
            'spav': True,
            # 'l_soc': True,
            'l_ss': True
        }
        fchanges.append(('set_inpchanges', {'change_dict': changes_dict}))

        if self.ctx.wf_dict['kmesh_force_theorem'] is not None:
            kmesh = KpointsData()
            kmesh.set_kpoints(monkhorst_pack(self.ctx.wf_dict['kmesh_force_theorem']))
            kmesh.store()
            fchanges.append(('set_kpointsdata', {
                'kpointsdata_uuid': kmesh.uuid,
                'switch': True,
                'kpoint_type': 'mesh'
            }))

        # change beta parameter
        for label, beta in self.ctx.wf_dict['beta'].items():
            fchanges.append(('set_atomgroup_label', {
                'attributedict': {
                    'nocoParams': {
                        'beta': beta
                    }
                },
                'atom_label': label
            }))

        # switch off SOC on an atom specie
        for atom_label in self.ctx.wf_dict['soc_off']:
            fchanges.append(('set_species_label', {
                'atom_label': atom_label,
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                },
            }))

        if fchanges:  # change inp.xml file
            fleurmode = FleurinpModifier(fleurin)
            try:
                fleurmode.add_task_list(fchanges)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply inpxml_changes'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

            # validate?
            try:
                fleurmode.show(display=False, validate=True)
            except etree.DocumentInvalid:
                error = ('ERROR: input, user wanted inp.xml changes did not validate')
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_FILE
            except ValueError as exc:
                error = ('ERROR: input, user wanted inp.xml changes could not be applied.'
                         f'The following error was raised {exc}')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

            # apply
            out = fleurmode.freeze()
            self.ctx.fleurinp = out
        else:  # otherwise do not change the inp.xml
            self.ctx.fleurinp = fleurin

        return

    def force_after_scf(self):
        '''
        This routine uses the force theorem to calculate energies dispersion of
        spin spirals. The force theorem calculations implemented into the FLEUR
        code. Hence a single iteration FLEUR input file having <forceTheorem> tag
        has to be created and submitted.
        '''
        calc = self.ctx.reference

        if not calc.is_finished_ok:
            message = ('The reference SCF calculation was not successful.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        try:
            outpara_node = calc.outputs.output_scf_wc_para
        except NotExistent:
            message = ('The reference SCF calculation failed, no scf output node.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        outpara = outpara_node.get_dict()

        if 'total_energy' not in outpara:
            message = ('Did not manage to extract float total energy from the reference SCF calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder of the reference calculation
        pk_last = 0
        scf_ref_node = load_node(calc.pk)
        for i in scf_ref_node.called:
            if i.node_type == 'process.workflow.workchain.WorkChainNode.':
                if i.process_class is FleurBaseWorkChain:
                    if pk_last < i.pk:
                        pk_last = i.pk
        try:
            remote = load_node(pk_last).outputs.remote_folder
        except AttributeError:
            message = ('Found no remote folder of the reference scf calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_NOREMOTE

        label = 'Force_theorem_calculation'
        description = 'This is a force theorem calculation for all SQA'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def force_wo_scf(self):
        """
        Submit FLEUR force theorem calculation using input remote
        """
        self.report('INFO: run Force theorem calculations')

        status = self.change_fleurinp()
        if status:
            return status

        fleurin = self.ctx.fleurinp

        # Do not copy mixing_history* files from the parent
        settings = {'remove_from_remotecopy_list': ['mixing_history*']}

        # Retrieve remote folder from the inputs
        remote = self.inputs.remote

        label = 'DMI_force_theorem'
        description = 'The is the force theorem calculation for DMI energy.'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        return ToContext(f_t=future)

    def get_results(self):
        """
        Generates results of the workchain.
        """
        mae_thetas = []
        mae_phis = []
        num_ang = 0
        q_vectors = []

        try:
            calculation = self.ctx.f_t
            if not calculation.is_finished_ok:
                message = f'ERROR: Force theorem Fleur calculation failed somehow it has exit status {calculation.exit_status}'
                self.control_end_wc(message)
                return self.exit_codes.ERROR_FORCE_THEOREM_FAILED
        except AttributeError:
            message = 'ERROR: Something went wrong I do not have a force theorem Fleur calculation'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        try:
            out_dict = calculation.outputs.output_parameters.dict
            h_so = out_dict.dmi_force_so_h_so
            mae_thetas = out_dict.dmi_force_theta
            mae_phis = out_dict.dmi_force_phi
            num_ang = out_dict.dmi_force_angles
            q_vectors = [self.ctx.wf_dict['q_vectors'][x - 1] for x in out_dict.dmi_force_so_q]
            e_u = out_dict.dmi_force_units
        except AttributeError:
            message = ('Did not manage to read evSum or energy units after FT calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        if not isinstance(h_so, list):
            message = ('Did not manage to read evSum or energy units after FT calculation.')
            self.control_end_wc(message)
            return self.exit_codes.ERROR_FORCE_THEOREM_FAILED

        if e_u in ['htr', 'Htr']:
            h_so = np.array(h_so) * HTR_TO_EV
            h_so = h_so.tolist()

        self.ctx.h_so = h_so
        self.ctx.q_vectors = q_vectors
        self.ctx.mae_thetas = mae_thetas
        self.ctx.mae_phis = mae_phis
        self.ctx.num_ang = num_ang

    def return_results(self):
        """
        This function outputs results of the wc
        """
        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            # 'initial_structure': self.inputs.structure.uuid,
            'is_it_force_theorem': True,
            'soc_energies': self.ctx.h_so,
            'q_vectors': self.ctx.q_vectors,
            'theta': self.ctx.mae_thetas,
            'phi': self.ctx.mae_phis,
            'angles': self.ctx.num_ang - 1,
            'energy_units': 'eV',
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors,
        }

        out = save_output_node(Dict(dict=out))
        self.out('out', out)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.report(errormsg)  # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def save_output_node(out):
    """
    This calcfunction saves the out dict in the db
    """
    out_wc = out.clone()
    return out_wc



# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit
from aiida.plugins import WorkflowFactory


fleur_corehole_wc = WorkflowFactory('fleur.corehole')
struc = load_node(STRUCTURE_PK)
flapw_para = load_node(PARAMETERS_PK)
fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)

options = Dict(dict={'resources': {'num_machines': 2, 'num_mpiprocs_per_machine': 24},
                         'queue_name': '',
                         'custom_scheduler_commands': '',
                         'max_wallclock_seconds':  60*60})

wf_para_corehole = Dict(dict={u'atoms': [u'Be'], #[u'all'],
  u'supercell_size': [2, 2, 2], u'corelevel': ['1s'], #[u'all'],
  u'hole_charge': 1.0, u'magnetic': True, u'method': u'valence'})

# launch workflow
dos = submit(fleur_corehole_wc, wf_parameters=wf_para_corehole, structure=struc,
             calc_parameters=flapw_para, options=options,
             fleur=fleur, inpgen=inpgen, label='test core hole wc',
             description='fleur_corehole test')



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module you find the workflow 'FleurStrainWorkChain' for the calculation of
of deformation potential
"""

import numpy as np

from aiida.plugins import DataFactory
from aiida.orm import Code, load_node
from aiida.orm import Float, StructureData, Dict
from aiida.engine import WorkChain, ToContext  # ,Outputs
from aiida.engine import calcfunction as cf

from aiida_fleur.tools.StructureData_util import rescale, rescale_nowf, is_structure
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf_util import check_eos_energies

# pylint: disable=invalid-name
FleurInpData = DataFactory('fleur.fleurinp')
# pylint: enable=invalid-name


class FleurStrainWorkChain(WorkChain):
    """
    This workflow calculates the deformation potential a structure = -BdEg/dP = d(Eg)/d(ln(V)).
    Calculates several unit cells with different volumes.
    A Birch_Murnaghan  equation of states fit determines the Bulk modulus(B) and the
    ground-state volume of the cell.

    :params wf_parameters: Dict node, optional 'wf_parameters', protocol specifying parameter dict
    :params structure: StructureData node, 'structure' crystal structure
    :params calc_parameters: Dict node, optional 'calc_parameters' parameters for inpgen
    :params inpgen: Code node,
    :params fleur: Code node,


    :return output_strain_wc_para: Dict node, contains relevant output information.
                                about general succeed, fit results and so on.
    """

    _workflowversion = '0.3.6'

    _default_options = {
        'resources': {
            'num_machines': 1
        },
        'max_wallclock_seconds': 6 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    _wf_default = {
        'fleur_runmax': 4,
        'density_converged': 0.02,
        'itmax_per_run': 30,
        'inpxml_changes': [],
        'points': 3,
        'step': 0.02,
        'guess': 1.00
    }

    _scf_keys = ['fleur_runmax', 'density_converged', 'itmax_per_run', 'inpxml_changes']

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('structure', valid_type=StructureData, required=True)
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.input('inpgen', valid_type=Code, required=True)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('options', valid_type=Dict, required=False)
        spec.input('settings', valid_type=Dict, required=False)

        spec.outline(cls.start, cls.structures, cls.converge_scf, cls.return_results)

        spec.output('output_strain_wc_para', valid_type=Dict)

        # exit codes
        spec.exit_code(331,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Invalid code node specified, check inpgen and fleur code nodes.')

    def start(self):
        """
        check parameters, what conditions? complete?
        check input nodes
        """
        self.report(f'Started strain workflow version {self._workflowversion}')

        self.ctx.last_calc2 = None
        self.ctx.calcs = []
        self.ctx.calcs_future = []
        self.ctx.structures = []
        self.ctx.temp_calc = None
        self.ctx.structurs_uuids = []
        self.ctx.scalelist = []
        self.ctx.volume = []
        self.ctx.volume_peratom = {}
        self.ctx.org_volume = -1  # avoid div 0
        self.ctx.labels = []
        self.ctx.successful = True
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        # TODO get all successful from convergence, if all True this

        # initialize the dictionary using defaults if no wf parameters are given
        wf_default = self._wf_default
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        self.ctx.points = wf_dict.get('points', 3)
        self.ctx.step = wf_dict.get('step', 0.02)
        self.ctx.guess = wf_dict.get('guess', 1.00)
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)

        # initialize the dictionary using defaults if no options are given
        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        # Check if user gave valid inpgen and fleur executables
        inputs = self.inputs
        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for inpgen of FLEUR does not use the plugin fleur.inpgen')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

    def structures(self):
        """
        Creates structure data nodes with different Volume (lattice constants)
        """
        points = self.ctx.points
        step = self.ctx.step
        guess = self.ctx.guess
        startscale = guess - (points - 1) / 2 * step

        for point in range(points):
            self.ctx.scalelist.append(startscale + point * step)

        self.report(f'scaling factors which will be calculated:{self.ctx.scalelist}')
        self.ctx.org_volume = self.inputs.structure.get_cell_volume()
        self.ctx.structurs = strain_structures(self.inputs.structure, self.ctx.scalelist)

    def converge_scf(self):
        """
        Launch fleur_scfs from the generated structures
        """
        calcs = {}

        for i, struc in enumerate(self.ctx.structurs):
            inputs = self.get_inputs_scf()
            inputs['structure'] = struc
            natoms = len(struc.sites)
            label = str(self.ctx.scalelist[i])
            label_c = '|strain| fleur_scf_wc'
            description = f'|FleurStrainWorkChain|fleur_scf_wc|scale {label}, {i}'
            # inputs['label'] = label_c
            # inputs['description'] = description

            self.ctx.volume.append(struc.get_cell_volume())
            self.ctx.volume_peratom[label] = struc.get_cell_volume() / natoms
            self.ctx.structurs_uuids.append(struc.uuid)

            result = self.submit(FleurScfWorkChain, **inputs)
            self.ctx.labels.append(label)
            calcs[label] = result

        return ToContext(**calcs)

    def get_inputs_scf(self):
        """
        get and 'produce' the inputs for a scf-cycle
        """
        inputs = {}

        scf_wf_param = {}
        for key in self._scf_keys:
            scf_wf_param[key] = self.ctx.wf_dict.get(key)
        inputs['wf_parameters'] = scf_wf_param

        inputs['options'] = self.ctx.options

        try:
            calc_para = self.inputs.calc_parameters.get_dict()
        except AttributeError:
            calc_para = {}
        inputs['calc_parameters'] = calc_para

        inputs['inpgen'] = self.inputs.inpgen
        inputs['fleur'] = self.inputs.fleur

        inputs['options'] = Dict(dict=inputs['options'])
        inputs['wf_parameters'] = Dict(dict=inputs['wf_parameters'])
        inputs['calc_parameters'] = Dict(dict=inputs['calc_parameters'])

        return inputs

    def return_results(self):
        """
        Return the results of the calculations  (scf workchains) and do a
        polynomial fit
        """
        distancelist = []
        t_energylist = []
        t_energylist_peratom = []
        bandgaplist = []
        calc_uuids = []
        vol_peratom_success = []
        natoms = len(self.inputs.structure.sites)
        htr_to_ev = 27.21138602

        for label in self.ctx.labels:
            calc = self.ctx[label]

            if not calc.is_finished_ok:
                message = f'One SCF workflow was not successful: {label}'
                self.ctx.warnings.append(message)
                self.ctx.successful = False
                continue

            try:
                _ = calc.outputs.output_scf_wc_para
            except KeyError:
                message = f'One SCF workflow failed, no scf output node: {label}. I skip this one.'
                self.ctx.errors.append(message)
                self.ctx.successful = False
                continue

            outpara = calc.outputs.output_scf_wc_para.get_dict()

            t_e = outpara.get('total_energy', float('nan'))
            e_u = outpara.get('total_energy_units', 'eV')
            if e_u in ('Htr', 'htr'):
                t_e = t_e * htr_to_ev
            dis = outpara.get('distance_charge', float('nan'))
            dis_u = outpara.get('distance_charge_units')
            t_energylist.append(t_e)
            t_energylist_peratom.append(t_e / natoms)
            vol_peratom_success.append(self.ctx.volume_peratom[label])
            distancelist.append(dis)
            calc_uuid = outpara.get('last_calc_uuid')
            calc_uuids.append(calc_uuid)
            bandgaplist.append(load_node(calc_uuid).res.bandgap)

        en_array = np.array(t_energylist_peratom)
        vol_array = np.array(vol_peratom_success)
        eg_array = np.array(bandgaplist)
        vol_unitcell_array = vol_array * natoms

        if len(en_array):
            volume, bulk_modulus, bulk_deriv, residuals = birch_murnaghan_fit(en_array, vol_array)
            dprime = deformation_potential(vol_unitcell_array, eg_array)

            volumes = self.ctx.volume
            gs_scale = volume * natoms / self.ctx.org_volume
            if (volume * natoms < volumes[0]) or (volume * natoms > volumes[-1]):
                warn = ('Groundstate volume was not in the scaling range.')
                hint = f'Consider rerunning around point {gs_scale}'
                self.ctx.info.append(hint)
                self.ctx.warnings.append(warn)
        else:
            volumes = None
            gs_scale = None
            residuals = None
            volume = 0
            bulk_modulus = None
            bulk_deriv = None
            dprime = None

        # if (self.inputs.structure.get_extra('local_name')):
        #     local_name=self.inputs.structure.get_extra('local_name')
        # else:
        #     local_name=''

        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['material'] = self.inputs.structure.get_formula()
        outputnode_dict['kind_names'] = self.inputs.structure.get_kind_names()
        # outputnode_dict['local_name'] = local_name
        outputnode_dict['deformation_potential'] = dprime
        outputnode_dict['scaling'] = self.ctx.scalelist
        outputnode_dict['scaling_gs'] = gs_scale
        outputnode_dict['initial_structure'] = self.inputs.structure.uuid
        outputnode_dict['volume_gs'] = volume * natoms
        outputnode_dict['volumes'] = volumes
        outputnode_dict['volume_units'] = 'A^3'
        outputnode_dict['natoms'] = natoms
        outputnode_dict['total_energy'] = t_energylist
        outputnode_dict['total_energy_units'] = e_u
        outputnode_dict['bandgaps'] = bandgaplist
        outputnode_dict['structures'] = self.ctx.structurs_uuids
        outputnode_dict['calculations'] = calc_uuids
        outputnode_dict['distance_charge'] = distancelist
        outputnode_dict['distance_charge_units'] = dis_u
        outputnode_dict['nsteps'] = self.ctx.points
        # outputnode_dict['guess']= self.ctx.guess,
        outputnode_dict['stepsize'] = self.ctx.step
        outputnode_dict['residuals'] = residuals
        outputnode_dict['bulk_deriv'] = bulk_deriv
        outputnode_dict['bulk_modulus'] = bulk_modulus * 160.217733  # * echarge * 1.0e21,#GPa
        outputnode_dict['bulk_modulus_units'] = 'GPa'
        outputnode_dict['info'] = self.ctx.info
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['errors'] = self.ctx.errors

        if self.ctx.successful:
            self.report('Done, Strain calculation complete')
        else:
            self.report('Done, but something went wrong.... Probably some individual calculation failed or'
                        ' a scf-cycle did not reach the desired distance.')

        outputnode_t = Dict(dict=outputnode_dict)
        outdict = create_strain_result_node(outpara=outputnode_t)

        # create link to work-chain node
        for link_name, node in outdict.items():
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shut-down the work-chain. It will initialize the output nodes
        The shut-down of the work-chain will has to be done afterwards
        """
        self.ctx.successful = False
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_strain_result_node(**kwargs):
    """
    Pseudo wf, to create the right graph structure of AiiDA.
    This work function will create the output node in the database.
    It also connects the output_node to all nodes the information comes from.
    """
    for key, val in kwargs.items():
        if key == 'outpara':
            outpara = val
    outdict = {}
    outputnode = outpara.clone()
    outputnode.label = 'output_strain_wc_para'
    outputnode.description = ('Contains results and information of a FleurStrainWorkChain run.')
    outdict['output_strain_wc_para'] = outputnode

    return outdict


def strain_structures(inp_structure, scalelist):
    """
    Creates many re-scaled StructureData nodes out of a crystal structure.
    Keeps the provenance in the database.

    :param StructureData, a StructureData node (pk, sor uuid)
    :param scale-list, list of floats, scaling factors for the cell

    :returns: list of New StructureData nodes with rescalled structure, which are linked to input
              Structure
    """
    structure = is_structure(inp_structure)
    if not structure:
        # TODO: log something (test if it gets here at all)
        return None
    re_structures = []

    for scale in scalelist:
        structure_rescaled = rescale(structure, Float(scale))  # this is a wf
        re_structures.append(structure_rescaled)

    return re_structures


# pylint: disable=invalid-name
def birch_murnaghan_fit(energies, volumes):
    """
    least squares fit of a Birch-Murnaghan equation of state curve. From delta project
    containing in its columns the volumes in A^3/atom and energies in eV/atom
    # The following code is based on the source code of eos.py from the Atomic
    # Simulation Environment (ASE) <https://wiki.fysik.dtu.dk/ase/>.
    :params energies: list (numpy arrays!) of total energies eV/atom
    :params volumes: list (numpy arrays!) of volumes in A^3/atom

    #volume, bulk_modulus, bulk_deriv, residuals = Birch_Murnaghan_fit(data)
    """
    fitdata = np.polyfit(volumes[:]**(-2. / 3.), energies[:], 3, full=True)
    ssr = fitdata[1]
    sst = np.sum((energies[:] - np.average(energies[:]))**2.)

    residuals0 = ssr / sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3. / 2.)
            break

    if volume0 == 0:
        print('Error: No minimum could be found')
        return None

    derivV2 = 4. / 9. * x**5. * deriv2(x)
    derivV3 = (-20. / 9. * x**(13. / 2.) * deriv2(x) - 8. / 27. * x**(15. / 2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3. / 2.)
    bulk_deriv0 = -1 - x**(-3. / 2.) * derivV3 / derivV2

    return volume0, bulk_modulus0, bulk_deriv0, residuals0


def deformation_potential(volume, bandgap):
    """Calculate the deformation potential"""
    xs = np.log(volume)
    ys = bandgap
    dprime = (((np.mean(xs) * np.mean(ys)) - np.mean(xs * ys)) / ((np.mean(xs)**2) - np.mean(xs**2)))
    return dprime



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Here we collect IO routines and their utility, for writting certain things to files, or post process files.
For example collection of data or database evaluations, for other people.
"""
import warnings


def write_results_to_file(headerstring, data, destination='./outputfile', seperator='  ', transpose=True):
    """
    Writes data to a file

    param headerstring: string with information
    param data: 2D array (numpy,scipy) with data [colum1 colum2, ...]
    """

    with open(destination, 'w') as thefile:
        thefile.write(headerstring)
        datastring = ''
        #seperator = seperator  # '\t'
        if transpose:
            datat = data.transpose()
        else:
            datat = data
        for item in datat:
            itemstring = ''
            for value in item:
                if isinstance(value, str):
                    itemstring = itemstring + f'{value}{seperator}'
                else:
                    itemstring = itemstring + f'{float(value):0.8f}{seperator:s}'
            datastring = datastring + itemstring.strip() + '\n'
        thefile.write(datastring)
    #thefile.close()


def write_xps_spectra_datafile(nodes,
                               factors,
                               natomtypes_dict,
                               bindingenergies,
                               bindingenergies_ref,
                               xdata_spec,
                               ydata_spec,
                               ydata_single_all,
                               xdata_all,
                               ydata_all,
                               compound_info,
                               xdatalabel,
                               destination='./outputfile'):
    '''
    special file write routine. Writes theoretical spectra data from plot spectra to file

    uses write_results_to_file
    '''
    import numpy as np

    dataprep = [xdata_spec, ydata_spec] + ydata_single_all
    data = np.array(dataprep)  #, ydata_single_all])

    formulastring = compound_info, factors
    nodesinvolvedstring = nodes
    atomtypesstring = natomtypes_dict
    be_string = bindingenergies
    reference_be_string = bindingenergies_ref

    headstring = (
        '# Theoretical Data provided as is without warranty. Copyright 2017-2018 Forschungszentrum Juelich GmbH\n'
        '# produced at PGI-1 with the FLEUR code within the AiiDA framework (with aiida-fleur), MIT License, please cite: \n'
        '################# Data meta information #############\n')
    tempst = (
        '# System: {}\n# Nodes: {}\n# Elements and natomtypes: {}\n# Bindingenergies [eV], pos: {}\n# Reference Energies [eV] pos: {}\n'
        ''.format(formulastring, nodesinvolvedstring, atomtypesstring, be_string, reference_be_string))
    tempst1 = ''
    for label in xdatalabel:
        tempst1 = tempst1 + ' | ' + label
    tempst2 = (f'#####################  Data  ######################\n# Energy [eV] | Total intensity {tempst1}\n')

    headstring = headstring + tempst + tempst2

    print(f'Writting theoretical XPS data to file: {destination}')
    write_results_to_file(headstring, data, destination=destination, seperator='  ')


# example/test
#from plot_methods.plot_fleur_aiida import plot_spectra
#import plot_methods
#all_wc_BeTi_uuid = ['107b0727-15cf-4436-b614-79801cdadd8c', 'f8b12b23-0b71-45a1-9040-b51ccf379439']
#factors = [1,1]
#returnvalues = plot_spectra(all_wc_BeTi_uuid, factors=factors, energy_range=[109,112], energy_grid=0.2)
#write_xps_spectra_datafile(all_wc_BeTi_uuid, factors, *returnvalues, destination='./out.txt')


def compress_fleuroutxml(outxmlfilepath, dest_file_path=None, delete_eig=True, iterations_to_keep=None):
    """
    Compresses a fleur out.xml file by deleting certain things
    like eigenvalues tags and/or iterations from it

    :param outxmlfilepath: (absolut) file path
    :type outxmlfilepath: str
    :param dest_file_path: (absolut) for the compressed file to be saved, if no desitination file path is given the file is overriden in place (default)!
    :type dest_file_path: str, optional
    :param delete_eig:  if eigenvalues are deleted from file default is True
    :type delete_eig: boolean, optional
    :param iterations_to_keep: index of 'till' whihc iterations to be keep, i.e '-2' means only last two, '15' default (None) is keep all
    :type iterations_to_keep: int

     ###
     usage example:
     outxmldes = '/Users/broeder/test/FePt_out_test.xml'
     outxmlsrc = '/Users/broeder/test/FePt_out.xml'
     compress_fleuroutxml(outxmlsrc, dest_file_path=outxmldes, iterations_to_keep=14)
     compress_fleuroutxml(outxmlsrc, dest_file_path=outxmldes, iterations_to_keep=-1)
    """
    from masci_tools.util.xml.xml_setters_names import delete_tag
    from masci_tools.util.schema_dict_util import get_number_of_nodes
    from masci_tools.util.schema_dict_util import eval_simple_xpath
    from masci_tools.io.io_fleurxml import load_outxml
    from lxml import etree

    xmltree, schema_dict = load_outxml(outxmlfilepath)
    xpath_iteration = schema_dict.tag_xpath('iteration')

    # delete eigenvalues (all)
    if delete_eig:
        iteration_nodes = eval_simple_xpath(xmltree, schema_dict, 'iteration', list_return=True)
        for iteration in iteration_nodes:
            #The explicit conversion is done since delete_tag expects a ElementTree in masci-tools <= 0.6.2
            delete_tag(etree.ElementTree(iteration), schema_dict, 'eigenvalues')

    # delete certain iterations
    if iterations_to_keep is not None:
        root = xmltree.getroot()
        n_iters = get_number_of_nodes(root, schema_dict, 'iteration')
        print(f'Found {n_iters} iterations')
        if iterations_to_keep < 0:
            # the first element has 1 (not 0) in xpath expresions
            position_keep = n_iters + iterations_to_keep + 1
            delete_xpath = f'{xpath_iteration}[position()<{int(position_keep)}]'
        else:
            delete_xpath = f'{xpath_iteration}[position()>{int(iterations_to_keep)}]'

        if abs(iterations_to_keep) > n_iters:
            warnings.warn(
                'iterations_to_keep is larger then the number of iterations'
                ' in the given out.xml file, I keep all.', UserWarning)
        else:
            print(f'Deleting iterations under the xpath: {delete_xpath}')
            xmltree = delete_tag(xmltree, schema_dict, 'iteration', complex_xpath=delete_xpath)

    etree.indent(xmltree)  #Make sure the print looks nice

    if dest_file_path is None:
        dest_file_path = outxmlfilepath  # overwrite file
    if xmltree.getroot() is not None:  #otherwise write fails
        xmltree.write(dest_file_path, encoding='utf-8', pretty_print=True)
    else:
        raise ValueError('xmltree has no root..., I cannot write to proper xml')


# usage example
#outxml15 = '/Users/broeder/test/FePt_out_test.xml'
#outxml14 = '/Users/broeder/test/FePt_out.xml'
#compress_fleuroutxml(outxml14, dest_file_path=outxml15, iterations_to_keep=14)
#compress_fleuroutxml(outxml14, dest_file_path=outxml15, iterations_to_keep=-1)
